#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Build args parser for get_cazy_cazymes.py"""


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, _SubParsersAction, Action
from pathlib import Path
from typing import List, Optional

from cazomevolve.scripts import download_acc_genomes


class ValidateFormats(Action):
    """Check the user has provided valid file formats, and convert to ncbi-genome formats"""
    def __call__(self, parser, args, values, option_string=None):
        valid_formats = {
            'genbank': 'genbank',
            'fasta': 'fasta',
            'rm': 'rm',
            'features': 'features',
            'gff': 'gff',
            'protein': 'protein-fasta',
            'genpept': 'genpept',
            'wgs': 'wgs',
            'cdsfasta': 'cds-fasta',
            'rnafna': 'rna-fasta',
            'assemblyreport': 'assembly-report',
            'assemblystats': 'assembly-stats',
            'all': 'all',
        }
        parsed_values = []
        invalid = False
        for value in values:
            if value not in list(valid_formats.keys()):
                invalid = True
                raise ValueError(f'Invalid file format "{value}" provided. Accepted formats: {list(valid_formats.keys())}')

            else:
                parsed_values.append(valid_formats[value])
        if invalid:
            sys.exit(1)

        setattr(args, self.dest, ','.join(parsed_values))


class ValidateDb(Action):
    """Check the user has provided valid database options."""
    def __call__(self, parser, args, values, option_string=None):
        valid_formats = ['genbank', 'refseq']
        invalid = False
        if values[0] not in valid_formats:
            invalid = True
            raise ValueError(f'Invalid database "{values[0]}" provided. Accepted formats: {valid_formats}')
        if invalid:
            sys.exit(1)
        setattr(args, self.dest, values[0])


class ValidateLevels(Action):
    """Check the user has provided valid assembly levels."""
    def __call__(self, parser, args, values, option_string=None):
        valid_formats = ['all', 'complete', 'chromosome', 'scaffold', 'contig']
        invalid = False
        for value in values:
            if value not in valid_formats:
                invalid = True
                raise ValueError(f'Invalid assembly level "{value}" provided. Accepted formats: {list(valid_formats.keys())}')
        if invalid:
            sys.exit(1)
        setattr(args, self.dest, ','.join(values))


def build_parser(
    subps: _SubParsersAction, parents: Optional[List[ArgumentParser]] = None
) -> None:
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = subps.add_parser(
        "download_acc_genomes", formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "accessions",
        type=str,
        help="Path to file listing the accessions, with a unique genome accession per row",
    )
    parser.add_argument(
        "outdir",
        type=str,
        help="output directory to write out the genomes to",
    )
    parser.add_argument(
        "file_opts",
        metavar="File options",
        nargs='+',
        action=ValidateFormats,
        choices=['genbank', 'fasta', 'rm', 'features', 'gff', 'protein', 'genpept', 'wgs', 'cdsfasta', 'rnafna', 'rnafasta', 'assemblyreport', 'assemblystats', 'all'],
        type=str,
        help=(
            "A space-separated list of file formats to download. For example: 'fasta assemblyreport'.\nChoose from:\n"
            "['genbank', 'fasta', 'rm', 'features', 'gff', 'protein', 'genpept', 'wgs', 'cdsfasta', 'rnafna', 'rnafasta', 'assemblyreport', 'assemblystats', 'all']"
        ),
    )
    parser.add_argument(
        "database",
        metavar='NCBI Database',
        action=ValidateDb,
        choices=['genbank', 'refseq'],
        nargs=1,
        type=str,
        help="Choose which NCBI db to get genomes from: refseq or genbank",
    )

    parser.add_argument(
        "-A",
        "--assembly_levels",
        nargs='+',
        action=ValidateLevels,
        choices=['all', 'complete', 'chromosome', 'scaffold', 'contig'],
        type=str,
        default='all',
        help=(
            "Space-separated list of assembly levels of genomes to download. Default='all'. Can provide multiple levels.\n"
            "Accepted = ['all', 'complete', 'chromosome', 'scaffold', 'contig']"
        ),
    )

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force file over writting",
    )
    parser.add_argument(
        "-n",
        "--nodelete",
        dest="nodelete",
        action="store_true",
        default=False,
        help="enable/disable deletion of exisiting files",
    )

    parser.set_defaults(func=download_acc_genomes.main)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Single entry point to explore CAZomes.

For full cusomisation of the operation, import the cazomevolve.cazome.explore modules 
into a jupyter notebook.
"""


import json
import logging
import pandas as pd
import sys

from copy import copy
from typing import List, Optional
from saintBioutils.utilities.file_io import make_output_directory

# loading and parsing data
from cazomevolve.cazome.explore.parse_data import (
    load_fgp_data,
    load_tax_data,
    add_tax_data_from_tax_df,
    add_tax_column_from_row_index,
)

# functions for exploring the sizes of CAZomes
from cazomevolve.cazome.explore.cazome_sizes import (
    calc_proteome_representation,
    count_items_in_cazome,
    get_proteome_sizes,
    count_cazyme_fam_ratio,
)

# explore the frequency of CAZymes per CAZy class
from cazomevolve.cazome.explore.cazy_classes import calculate_class_sizes

# explore the frequencies of CAZy families and identify the co-cazome
from cazomevolve.cazome.explore.cazy_families import (
    build_fam_freq_df,
    build_row_colours,
    build_family_clustermap,
    identify_core_cazome,
    build_fam_mean_freq_df,
    get_group_specific_fams,
)

# functions to identify and explore CAZy families that are always present together
from cazomevolve.cazome.explore.cooccurring_families import (
    calc_cooccuring_fam_freqs,
    add_to_upsetplot_membership,
    build_upsetplot,
    get_upsetplot_grps,
    add_upsetplot_grp_freqs,
    build_upsetplot_matrix,
)

# functions to perform PCA
from cazomevolve.cazome.explore.pca import (
    perform_pca,
    plot_explained_variance,
    plot_scree,
    plot_pca,
    plot_loadings,
)
from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    # make parent output directory
    if str(args.output_dir.parent) != ".":
        make_output_directory(args.output_dir, args.force, args.nodelete)

    col_names = []
    if args.kingdom:
        col_names.append('Kingdom')
    if args.phylum:
        col_names.append('Phylum')
    if args.tax_class:
        col_names.append('Class')
    if args.tax_order:
        col_names.append('Order')
    if args.tax_family:
        col_names.append('Family')
    if args.genus:
        col_names.append('Genus')
    if args.species:
        col_names.append('Species')

    if len(col_names) == 0:
        logger.warning("Must specify at least one rank of lineage. These are the taxonomic ranks listed in the Taxonomy CSV file")
        sys.exit(1)

    fgp_df = load_data(args)

    compare_cazome_sizes(fgp_df, args)

    compare_cazy_classes(fgp_df, args)

    fam_freq_df, fam_freq_df_ggs, all_families = compare_cazy_families(fgp_df, args)

    compare_core_cazomes(fam_freq_df, fam_freq_df_ggs, all_families, args)

    find_always_cooccurring_families(fam_freq_df, fam_freq_df_ggs, all_families, args)

    closing_message('Explore CAZomes', args)


def load_data(args):
    """Load in all data required for the analysis
    
    :param args: CLI args parser

    Return fgp_df, dataframe of cazy family, genome, protein id, and one column per tax rank
    """
    logger = logging.getLogger(__name__)

    # Load cazy family annotations
    fgp_df = load_fgp_data(args.fgp_file)
    logger.warning(
        f"Total CAZymes (i.e. the number of unique protein IDs): {len(set(fgp_df['Protein']))}"
    )

    # load tax data
    tax_df = load_tax_data(
        args.tax_csv_path,
        kingdom=args.kingdom,
        phylum=args.phylum,
        tax_class=args.tax_class,
        tax_order=args.tax_order,
        tax_family=args.tax_family,
        genus=args.genus,
        species=args.species,
    )

    # compile data into a single dataframe
    fgp_df = add_tax_data_from_tax_df(
        fgp_df,
        tax_df,
        genus=True,
        species=True,
    )

    return fgp_df


def compare_cazome_sizes(fgp_df, args):
    """Explore and compare the sizes of CAZomes by calculating:
    * The number of CAZymes per genome
    * The mean number of CAZymes per genome per genus
    * The proportion of the proteome represented by the CAZome
    * The mean proportion of the proteome represented by the CAZome

    The number of CAZymes is the number of unique protein IDs

    :param fgp_df: pandas df of cazy family, genome, protein id, and one col per tax rank
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "cazome_size"
    make_output_directory(outdir, force=True, nodelete=True)
    outpath = outdir / "cazome_sizes.csv"

    logger.warning(f"Examining {len(set(fgp_df['Genome']))} genomes")

    # count number of CAZymes
    cazome_sizes_dict, cazome_sizes_df = count_items_in_cazome(
        fgp_df,
        'Protein',
        args.group_by,
        round_by=args.round_by,
    )

    # Count number of CAZy families
    cazome_fam_dict, cazome_fams_df = count_items_in_cazome(
        fgp_df,
        'Family',
        args.group_by,
        round_by=args.round_by,
    )

    # Calculate CAZyme to CAZy family ratio
    cazome_ratio_dict, cazome_ratio_df = count_cazyme_fam_ratio(
        fgp_df,
        args.group_by,
        round_by=args.round_by,
    )

    # Calculate proteome sizes
    if args.proteome_dir is not None:
        proteome_dict = get_proteome_sizes(args.proteome_dir, fgp_df, args.group_by)
        total_proteins = 0
        for genus in proteome_dict:
            for genome in proteome_dict[genus]:
                total_proteins += proteome_dict[genus][genome]['numOfProteins']
        logger.warning(f"Total number of proteins across all genomes: {total_proteins}")

        # calculate precentage of the proteome that is made up of the cazome
        proteome_perc_df = calc_proteome_representation(
            proteome_dict,
            cazome_sizes_dict,
            args.group_by,
            round_by=args.round_by,
        )

        # combine into a single df
        all_df = pd.concat([proteome_perc_df, cazome_sizes_df, cazome_fams_df, cazome_ratio_df], axis=1, join='inner')

    else:
        all_df = pd.concat([cazome_sizes_df, cazome_fams_df, cazome_ratio_df], axis=1, join='inner')

    logger.warning(
        f"Writing out dataframe summarising CAZome sizes (with means and SD per {args.group_by})\n"
        f"to: {outpath}"
    )
    all_df.to_csv(outpath)


def compare_cazy_classes(fgp_df, args):
    """Compare the composition of the CAZy classes.
    
    Compare the number of CAZymes (i.e. unique protein IDs) per CAZy class 
    and the percentage of the CAZome encapsulated by each CAZy class.
    
    :param fgp_df: pandas df of cazy family, genome, protein id, and one col per tax rank
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "cazy_classes"
    make_output_directory(outdir, force=True, nodelete=True)
    outpath = outdir / "cazy_classes.csv"

    class_df, class_size_dict = calculate_class_sizes(
        fgp_df,
        args.group_by,
        round_by=args.round_by,
    )

    logger.warning(f"Writing out dataframe of CAZy class frequencies to {outpath}")
    class_df.to_csv(outpath)


def compare_cazy_families(fgp_df, args):
    """Compare the composition of the Families classes.
    
    Compare the number of CAZymes (i.e. unique protein IDs) per CAZy families
    
    :param fgp_df: pandas df of cazy family, genome, protein id, and one col per tax rank
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "cazy_families"
    make_output_directory(outdir, force=True, nodelete=True)
    outpath = outdir / "cazy_family_frequencies.csv"

    fam_freq_df = build_fam_freq_df(
        fgp_df, [args.group_by]
    )

    logger.warning(f"Writing out dataframe of CAZy family frequencies per genome to {outpath}")
    fam_freq_df.to_csv(outpath)

    # build a clustermap

    # index the taxonomy data and genome (ggs=genome_genus_species)
    fam_freq_df_ggs = copy(fam_freq_df)  # so does not alter fam_freq_df
    index = ['Genome']
    if args.kingdom:
        index.append('Kingdom')
    if args.phylum:
        index.append('Phylum')
    if args.tax_class:
        index.append('Class')
    if args.tax_order:
        index.append('Order')
    if args.tax_family:
        index.append('Family')
    if args.genus:
        index.append('Genus')
    if args.species:
        index.append('Species')

    fam_freq_df_ggs = fam_freq_df_ggs.set_index(index)
    # define a colour scheme to colour code rows by genus
    fam_freq_df_ggs[args.group_by] = list(fam_freq_df[args.group_by])  # add column to use for colour scheme, is removed
    fam_freq_genus_row_colours, fam_g_lut = build_row_colours(fam_freq_df_ggs, args.group_by, 'Set2')

    for file_format in args.formats:
        outpath_cm = outdir / f"cazy_family_clustermap.{file_format}"
        logger.warning(
            f"Writing out clustermap of CAZy family frequencies in {file_format} format to:\n"
            f"{outpath_cm}"
        )
        build_family_clustermap(
            fam_freq_df_ggs,
            row_colours=fam_freq_genus_row_colours,
            fig_size=(len(fam_freq_df_ggs.columns)*0.4, len(fam_freq_df_ggs)*0.4),
            file_path=outpath_cm,
            file_format=format,
            lut=fam_g_lut,
            legend_title=args.group_by,
            dendrogram_ratio=(0.2, 0.05),
            title_fontsize=28,
            legend_fontsize=24,
            cbar_pos=(0, 0.95, 0.05, 0.05),
            show=args.show_plots,
        )

    # find group specific families
    outpath_dic = outdir / f"{args.group_by}_specific_cazy_families.json"
    all_families = list(fam_freq_df.columns)[3:]
    unique_grp_fams, group_fams = get_group_specific_fams(fam_freq_df, args.group_by, all_families)
    for grp in unique_grp_fams:
        unique_grp_fams[grp] = list(unique_grp_fams[grp])
    logger.warning(f"Writing out {args.group_by} specific CAZy families to {outpath_dic}")
    with open(outpath_dic, "w") as fh:
        json.dump(unique_grp_fams, fh)

    return fam_freq_df, fam_freq_df_ggs, all_families


def compare_core_cazomes(fam_freq_df, fam_freq_df_ggs, all_families, args):
    """Comprae core CAZome across the entire data set and per group
    
    :param fam_freq_df: dataframe of CAZy fam freqs, genome per row, fam per column, and tax columns
    :param fam_Freq_df_ggs: same as fam_freq_df by tax data and genome are the row index
    :param all_families: list of all CAZy families in the genomes
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "core_cazome"
    make_output_directory(outdir, force=True, nodelete=True)
    outpath = outdir / "core_cazome.txt"

    index = ['Genome']
    if args.kingdom:
        index.append('Kingdom')
    if args.phylum:
        index.append('Phylum')
    if args.tax_class:
        index.append('Class')
    if args.tax_order:
        index.append('Order')
    if args.tax_family:
        index.append('Family')
    if args.genus:
        index.append('Genus')
    if args.species:
        index.append('Species')

    try:
        fam_freq_df_ggs = fam_freq_df_ggs.set_index(index)
    except KeyError:
        pass
    
    core_cazome = identify_core_cazome(fam_freq_df_ggs)
    with open(outpath, "w") as fh:
        fh.write(
            f"Total families across all genomes: {len(all_families)}\nThe core CAZyme families are:"
        )
        for fam in core_cazome:
            fh.write(f"{fam}\n")
    
    logger.warning(f"The core CAZome: {core_cazome}")
    logger.warning(f"The core CAZome was written to {outpath}")

    # get fam freqs by group
    core_cazome_df = fam_freq_df_ggs[core_cazome]
    core_cazome_df_grped = add_tax_column_from_row_index(core_cazome_df, args.group_by, 1)
    core_cazome_fggf_df, core_cazome_mean_freq_df = build_fam_mean_freq_df(
        core_cazome_df_grped,
        args.group_by,
        round_by=args.round_by,
    )
    outpath_df = outdir / "core_cazome_freqs.csv"
    logger.warning(
        "Writing the mean (and SD) frequency per family in the core "
        f"CAZome per {args.group_by} to {outpath_df}"
    )
    core_cazome_mean_freq_df.to_csv(outpath_df)


def find_always_cooccurring_families(fam_freq_df, fam_freq_df_ggs, all_families, args):
    """Comprae core CAZome across the entire data set and per group
    
    :param fam_freq_df: dataframe of CAZy fam freqs, genome per row, fam per column, and tax columns
    :param fam_Freq_df_ggs: same as fam_freq_df by tax data and genome are the row index
    :param all_families: list of all CAZy families in the genomes
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "cooccurring_families"
    make_output_directory(outdir, force=True, nodelete=True)
    outpath_all = outdir / "cooccurring_families.txt"
    outpath_grp = outdir / f"{args.group_by}_cooccurring_families.txt"

    cooccurring_fams_dict = calc_cooccuring_fam_freqs(
        fam_freq_df,
        list(all_families),
        exclude_core_cazome=False,
    )
    with open(outpath_all, "w") as fh:
        fh.write(str(cooccurring_fams_dict))

    grp_cooccuring_fams = {}  # {genus: cooccurring_fams_d
    for grp in set(fam_freq_df[args.group_by]):
        grp_fam_freq_df = fam_freq_df[fam_freq_df[args.group_by] == grp]
        grp_cooccurring_fams_dict = calc_cooccuring_fam_freqs(
            grp_fam_freq_df,
            list(all_families),
            exclude_core_cazome=False,
        )
        grp_cooccuring_fams[grp] = grp_cooccurring_fams_dict
    with open(outpath_grp, "w") as fh:
        fh.write(str(grp_cooccuring_fams))

    upsetplot_membership = []
    upsetplot_membership = add_to_upsetplot_membership(upsetplot_membership, cooccurring_fams_dict)

    for genus in grp_cooccuring_fams:
        upsetplot_membership = add_to_upsetplot_membership(
            upsetplot_membership,
            grp_cooccuring_fams[genus],
        )
    
    for file_format in args.formats:
        upset_path = outdir / f"cooccurring_fams_upsetplot.{file_format}"
        logger.warning(
            f"Writing out Upset plot of always co-occurring CAZy families in {file_format} format to:\n"
            f"{upset_path}"
        )
        build_upsetplot(
            upsetplot_membership,
            file_path=upset_path,
        )

    # calculate frequencies
    upset_plot_groups = get_upsetplot_grps(upsetplot_membership)
    cooccurring_grp_freq_data = []
    cooccurring_grp_freq_data = add_upsetplot_grp_freqs(
        upset_plot_groups,
        cooccurring_grp_freq_data,
        grp_cooccuring_fams,
        args.group_by,
        grp_sep=True,
        include_none=True,
    )
    output_path = outdir / "cooccurring_fam_freqs.csv"
    logger.warning(f"Writing out frequencies of always co-occurring families to {output_path}")
    cooccurring_fams_freq_df = build_upsetplot_matrix(
        cooccurring_grp_freq_data,
        args.group_by,
        file_path=output_path,
    )


def run_pca(fam_freq_df, fam_freq_df_ggs, all_families, args):
    """Run principal component analysis on CAZy family frequencies
    
    :param fam_freq_df: dataframe of CAZy fam freqs, genome per row, fam per column, and tax columns
    :param fam_Freq_df_ggs: same as fam_freq_df by tax data and genome are the row index
    :param all_families: list of all CAZy families in the genomes
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)
    outdir = args.output_dir / "pca"
    make_output_directory(outdir, force=True, nodelete=True)
    index = []
    if args.kingdom:
        index.append('Kingdom')
    if args.phylum:
        index.append('Phylum')
    if args.tax_class:
        index.append('Class')
    if args.tax_order:
        index.append('Order')
    if args.tax_family:
        index.append('Family')
    if args.genus:
        index.append('Genus')
    if args.species:
        index.append('Species')

    for col in index:
        try:
            fam_freq_df_ggs = fam_freq_df_ggs.drop(col, axis=1)
        except KeyError:
            pass
    
    num_of_components = len(fam_freq_df_ggs.columns)
    try:
        pca, X_scaled = perform_pca(fam_freq_df_ggs, num_of_components)
    except Exception:
        num_of_components = len(fam_freq_df_ggs)
        pca, X_scaled = perform_pca(fam_freq_df_ggs, num_of_components)
    
    logger.warning(
        f"{round(pca.explained_variance_ratio_.sum() * 100, 2)}% "
        "of the variance in the data set was catpured by the PCA"
    )

    for file_format in args.formats:
        out_cumvar = outdir / f"pca_explained_variance.{file_format}"
        logger.warning(
            f"Writing out plot of explained cumulative variance from the PCA in {file_format} to \n:"
            f"{out_cumvar}"
        ) 
        cumExpVar = plot_explained_variance(
            pca,
            num_of_components,
            file_path=out_cumvar,
            file_format=file_format,
            show=args.show_plots,
        )

        out_scree = outdir / f"pca_scree_plot.{file_format}"
        logger.warning(
            f"Writing out scree plot from the PCA in {file_format} to \n:"
            f"{out_scree}"
        )
        plot_scree(
            pca,
            nComp=10,
            file_format=file_format,
            file_path=out_scree,
            show=args.show_plots,
        )

    # plot PCS 1-4
    plot_pcs((1,2), fam_freq_df_ggs, pca, X_scaled, outdir, args)
    plot_pcs((1,3), fam_freq_df_ggs, pca, X_scaled, outdir, args)
    plot_pcs((1,4), fam_freq_df_ggs, pca, X_scaled, outdir, args)
    plot_pcs((2,3), fam_freq_df_ggs, pca, X_scaled, outdir, args)
    plot_pcs((2,4), fam_freq_df_ggs, pca, X_scaled, outdir, args)
    plot_pcs((3,4), fam_freq_df_ggs, pca, X_scaled, outdir, args)


def plot_pcs(pc_pair, fam_freq_df_ggs, pca, X_scaled, outdir, args):
    """Project genomes onto PCS, plot loadings
    
    :param pc_pair: tuple, number of PCs to plot
    :param outdir: path to output dir
    :param args: CLI args parser
    """
    logger = logging.getLogger(__name__)

    pc_outdir = outdir / f"PC{pc_pair[0]}-vs-PC{pc_pair[1]}"
    make_output_directory(pc_outdir, force=True, nodelete=True)

    logger.warning(f"PC{pc_pair[0]} vs PC{pc_pair[1]} - plotting scatter plot")

    for file_format in args.formats:
        out = pc_outdir / f'pca_pc{pc_pair[0]}_vs_pc{pc_pair[1]}-{args.group_by}.{file_format}'
        g = plot_pca(
            pca,
            X_scaled,
            fam_freq_df_ggs,
            pc_pair[0],
            pc_pair[1],
            args.group_by,
            style=args.group_by,
            figsize=(15,15),
            file_path=out,
            file_format=file_format,
            show=args.show_plots,
        );

    logger.warning(f"PC{pc_pair[0]} vs PC{pc_pair[1]} - plotting loadings plot")
    for file_format in args.formats:
        out = pc_outdir / f'pca_pc{pc_pair[0]}_vs_pc{pc_pair[1]}-{args.group_by}-loadings.{file_format}'
        plot_loadings(
            pca,
            fam_freq_df_ggs,
            pc_pair[0],
            pc_pair[1],
            threshold=0.3,
            figsize=(10,10),
            file_path=out,
            font_size=11,
            file_format=file_format,
            show=args.show_plots,
        );



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Explore the sizes of CAZy family populations per genome"""


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from matplotlib.patches import Patch

from tqdm import tqdm


def build_fam_freq_df(gfp_df, tax_ranks):
    """Build matrix of fam freq per genome
    
    Each row represents a genome, each column a CAZy family
    
    :param gfp_df: pandas df - tab delimit list of ['Family', 'Genome', 'Protein', 'tax1', 'tax2'...]
    :param tax_ranks: list of tax ranks to include the matrix, one column generated per rank
        Must match columns names in gfp_df, e.g. ['Genus', 'Species']
    
    Return matrix as pandas df
    """
    # identify all families present in the dataset
    all_families = set(gfp_df['Family'])
    all_families = list(all_families)
    all_families.sort()
    print(f"The dataset contains {len(all_families)} CAZy families")
    
    # identify all genomes i the dataset
    all_genomes = set(gfp_df['Genome'])
    
    # define column names
    col_names = ['Genome']
    
    for rank in tax_ranks:
        col_names.append(rank)
        
    for fam in all_families:
        col_names.append(fam)
        
    # gather fam freq data per genome
    fam_df_data = []

    for genome in tqdm(all_genomes, desc="Counting fam frequencies"):
        row_data = [genome]

        # get tax data
        for rank in tax_ranks:
            row_data.append(gfp_df[gfp_df['Genome'] == genome].iloc[0][rank])

        # gather all genome rows
        g_rows = gfp_df[gfp_df['Genome'] == genome]

        # count number of proteins in the family
        for fam in all_families:
            fam_rows = g_rows[g_rows['Family'] == fam]
            fam_freq = len(set(fam_rows['Protein']))
            row_data.append(fam_freq)

        fam_df_data.append(row_data)

    fam_freq_df = pd.DataFrame(fam_df_data, columns=col_names)
    
    return fam_freq_df


def build_row_colours(df, grp, palette):
    """Build map of colour to member of grp (e.g. genus)

    The dataframe that is parsed to `build_row_colours()` must be the dataframe that is used to 
    generate a clustermap, otherwise Seaborn will not be able to map the row oclours correctly 
    and no row colours will be produced.

    The dataframe used to generate the clustermap when passed to the function, must include the 
    column to be used to define the row colours, e.g. a 'Genus' column. This column (named by `grp`)
    is removed within the function.
    
    :param df: matrix of genome x fam, with fam freq
    :param grp: str, name of col to map colour scheme onto, e.g. 'Genus' or 'Species'
    :param palette: str, name of seaborn colour scheme to use, e.g. Set1
    
    Return map and lut
    """
    series = df.pop(grp)
    lut = dict(zip(
        series.unique(),
        sns.color_palette(palette, n_colors=len(list(series.unique())))
    ))
    row_colours = series.map(lut)
    
    return row_colours, lut


def build_family_clustermap(
    df,
    row_colours=None,
    fig_size=None,
    file_path=None,
    file_format='png',
    font_scale=1,
    dpi=300,
    dendrogram_ratio=0.2,
    lut=None,
    legend_title='',
    title_fontsize='2',
    legend_fontsize='2',
    bbox_to_anchor=(1,1),
    cmap=sns.cubehelix_palette(dark=1, light=0, reverse=True, as_cmap=True),
    cbar_pos=(0.02, 0.8, 0.05, 0.18),
    show=False,
):
    """Build a clustermap of the CAZy family frequencies per genome
    
    :param df: df of CAZy family frequencies per genome
    :param row_colours: pandas map - used to define additional row colours. or list of maps for 
        multiple sets of row colours. If None, additional row colours are not plotted
    :param fig_size: tuple (width, height) of final figure. If None, decided by Seaborn
    :param file_path: path to save image to. If None, the figure is not written to a file
    :param file_format: str, file format to save figure to. Default 'png'
    :param font_scale: int, scale text - use if text is overlapping. <1 to reduce 
        text size
    :param dpi: dpi of saved figure
    :param dendrogram_ratio: Proportion of the figure size devoted to the dendrograms.
        If a pair is given, they correspond to (row, col) ratios.
    :param lut: lut from generating colour scheme, add to include legend in the plot7
    :param legend_title: str, title of legend for row colours
    :title_fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        The font size of the legend's title.
    :legend_fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
    :param bbox_to_anchor: tuple, coordinates to place legend
    :param cmap: Seaborn cmap to be used for colour scheme of the heat/clustermap
    :param cbar_pos: from seaborn.clustermap, position and size of colour scale key/bar
        seaborn default=(0.02, 0.8, 0.05, 0.18) - left, bottom, width, height
    
    Return clustermap object
    """
    sns.set(font_scale=font_scale)
    
    fam_clustermap = sns.clustermap(
        df,
        cmap=cmap,
        figsize=fig_size,
        row_colors=row_colours,
        dendrogram_ratio=dendrogram_ratio,
        yticklabels=True,
        xticklabels=True,
        cbar_pos=cbar_pos,
    );
    
    if lut is not None:
        handles = [Patch(facecolor=lut[name]) for name in lut]
        plt.legend(
            handles,
            lut,
            title=legend_title,
            bbox_to_anchor=bbox_to_anchor,
            bbox_transform=plt.gcf().transFigure,
            loc='upper center',
            title_fontsize=title_fontsize,
            fontsize=legend_fontsize,
        )
    
    if show:
        plt.show();

    if file_path is not None:
        fam_clustermap.savefig(
            file_path,
            dpi=dpi,
            bbox_inches='tight',
            format=file_format,
        )

    return fam_clustermap


def build_family_clustermap_multi_legend(
    df,
    row_colours,
    luts,
    legend_titles,
    bbox_to_anchors,
    legend_cols=None,
    fig_size=None,
    file_path=None,
    file_format='png',
    font_scale=1,
    dpi=300,
    dendrogram_ratio=0.2,
    title_fontsize=2,
    legend_fontsize=2,
    cmap=sns.cubehelix_palette(dark=1, light=0, reverse=True, as_cmap=True),
    cbar_pos=(0.02, 0.8, 0.05, 0.18),
):
    """Build a clustermap of the CAZy family frequencies per genome
    
    :param df: df of CAZy family frequencies per genome
    :param row_colours: List of maps for multiple sets of row colours
    :param luts: list of luts, in same order as row_colours
    :param legend_titles: list of legend titles, in same order as luts and row_colours
    :param bbox_to_anchors: list of tuples, coordinates to place legends. One tuple per legend
    
    :param legend_cols: list of ints, number of cols to put in each legend. One int per legend
    :param fig_size: tuple (width, height) of final figure. If None, decided by Seaborn
    :param file_path: path to save image to. If None, the figure is not written to a file
    :param file_format: str, file format to save figure to. Default 'png'
    :param font_scale: int, scale text - use if text is overlapping. <1 to reduce 
        text size
    :param dpi: dpi of saved figure
    :param dendrogram_ratio: Proportion of the figure size devoted to the dendrograms.
        If a pair is given, they correspond to (row, col) ratios.
    :title_fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        The font size of the legend's title.
    :legend_fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
    :param cmap: Seaborn cmap to be used for colour scheme of the heat/clustermap
    :param cbar_pos: from seaborn.clustermap, position and size of colour scale key/bar
        seaborn default=(0.02, 0.8, 0.05, 0.18) - left, bottom, width, height
    
    Return clustermap object
    """
    if legend_cols is None:
        legend_cols = [1] * len(luts)
    
    sns.set(font_scale=font_scale)
    
    fam_clustermap = sns.clustermap(
        df,
        cmap=cmap,
        figsize=fig_size,
        row_colors=row_colours,
        dendrogram_ratio=dendrogram_ratio,
        yticklabels=True,
        xticklabels=True,
        cbar_pos=cbar_pos,
    );

    for i in range(len(luts)):
        if i == 0:
            lut = luts[i]
            labels = set(lut.keys())
            title = legend_titles[i]
            bbox_to_anchor = bbox_to_anchors[i]
            ncols = legend_cols[i]
            
            for label in labels:
                fam_clustermap.ax_row_dendrogram.bar(0, 0, color=lut[label], label=label, linewidth=0);
            l1 = fam_clustermap.ax_row_dendrogram.legend(
                title=title,
                loc="center",
                ncol=ncols,
                bbox_to_anchor=bbox_to_anchor,
                bbox_transform=plt.gcf().transFigure,
                title_fontsize=title_fontsize,
                fontsize=legend_fontsize,
            )   
            
        else:
            lut = luts[i]
            labels = set(lut.keys())
            title = legend_titles[i]
            bbox_to_anchor = bbox_to_anchors[i]
            ncols = legend_cols[i]
            handles = [Patch(facecolor=lut[name]) for name in lut]
            plt.legend(
                handles,
                lut,
                title=title,
                bbox_to_anchor=bbox_to_anchor,
                bbox_transform=plt.gcf().transFigure,
                loc='center',
                title_fontsize=title_fontsize,
                fontsize=legend_fontsize,
                ncol=ncols,
            )
        
    if file_path is not None:
        fam_clustermap.savefig(
            file_path,
            dpi=dpi,
            bbox_inches='tight',
            format=file_format,
        )

    return fam_clustermap


## For the core CAZome


def identify_core_cazome(df):
    """Identify families that are present in every genome
    
    :param df: pandas df, matrix where each row is a genome, and each column a CAZy family
    
    Return set of CAZy families"""
    core_cazome = set()

    for fam in tqdm(df.columns, desc="Identifying core CAZome"):
            if 0 not in set(df[fam]):
                core_cazome.add(fam)

    return core_cazome


def plot_fam_boxplot(
    df,
    font_scale=1,
    file_path=None,
    file_format='png',
    dpi=300,
    fig_size=None,
):
    """Plot a one-dimensional boxplot of the frequencies of the CAZy families in the df
    
    :param :param df: pandas df, matrix where each row is a genome, and each column a CAZy family
    :param font_scale: int, >1 increase font size, <1 to reduce font size
    :param file_path: path to save image to. If None, the figure is not written to a file
    :param file_format: str, file format to save figure to. Default 'png'
    :param dpi: dpi of saved figure
    :param fig_size: tuple, (width, height) mannually define fig_size
    
    Return nothing.
    """
    sns.set(font_scale=font_scale)
    
    if fig_size is not None:
        sns.set(rc={'figure.figsize':fig_size})
    
    # build long form df of [Fam, freq in a genome]
    long_form_data = []
    
    for i in range(len(df)):
        row = df.iloc[i]
        for fam in df.columns:
            row_data = [fam, row[fam]]
            long_form_data.append(row_data)
    
    long_form_df = pd.DataFrame(long_form_data, columns=["Family", "Frequency"])

    boxplot = sns.boxplot(x=long_form_df['Family'], y=long_form_df['Frequency'])
    
    if file_path is not None:
        boxplot.savefig(
            file_path,
            dpi=dpi,
            bbox_inches='tight',
            format=file_format,
        )


def build_fam_mean_freq_df(df, grp, round_by=None):
    """Build two dataframes of fam frequencies from a wide fam freq df
    
    DF 1: Family, tax rank (i.e. group), genome, freq
    DF 2: Family, tax rank (i.e. group), mean freq, sd freq
    
    :param df: pandas df, each row is a genome and each column a CAZy family
        and one column with tax rank listed (e.g. a 'Genus' column)
        and index includes the genomic accession
    :param grp: str, name of tax rank to group data by, and matches a name of one 
        of the columns in the dataframe (e.g. a 'Genus' column)
    :param round_by: int, number of decimal points to round by. If None, does not round
    
    Return two dataframes
    """
    families = list(df.columns)
    families.remove(grp)
    
    df_1_data = []  # Family, tax rank (i.e. group), genome, freq
    
    for i in tqdm(range(len(df)), desc="Building [fam, grp, genome, freq] df"):
        row = df.iloc[i]
        
        genome = row.name
        grp_name = row[grp]
        
        for fam in families:
            freq = row[fam]
            df_1_data.append([fam, grp_name, genome, freq])
            
    df_1 = pd.DataFrame(df_1_data, columns=['Family', grp, 'Genome', 'Frequency'])
    
    # build second df of Family, tax rank (i.e. group), mean freq, sd freq
    groups = set(df_1[grp])
    
    df_2_data = []
    for grp_name in tqdm(groups, desc="Building [Fam, grp, mean freq, sd freq] df"):
        grp_rows = df_1[df_1[grp] == grp_name]
        
        for fam in families:
            fam_grp_rows = grp_rows[grp_rows['Family'] == fam]
            mean = np.mean(fam_grp_rows['Frequency'])
            sd = np.std(fam_grp_rows['Frequency'])

            if round_by is not None:
                mean = round(mean, round_by)
                sd = round(sd, round_by)
            
            df_2_data.append([fam, grp_name, mean, sd])
        
    df_2 = pd.DataFrame(df_2_data, columns=['Family',grp,'MeanFreq','SdFreq'])
    
    return df_1, df_2


def get_group_specific_fams(fam_freq_df, group_by, all_families):
    """Identify families that are present in only one group
    
    The taxonomic information needs to be contained in the row names, use index_df() from cazomevolve
    
    :param fam_freq_df: df, rows=genomes, cols=fam freqs and column containing data to group
        genomes by, e.g. a 'Genus' column
    :param group_by: str, name of column to group genomes by
    :param all_families: list of CAZy families to analyse
    
    Return dict {group: {only unique fams}} and dict {group: {all fams}}
    """
    # Identify the families present in each group
    group_fams = {}  # {group: {fams}}

    # identify all fams in each group
    for ri in tqdm(range(len(fam_freq_df)), desc=f"Identifying fams in each {group_by}"):
        group = fam_freq_df.iloc[ri][group_by]

        try:
            group_fams[group]
        except KeyError:
            group_fams[group] = set()

        for fam in all_families:
            if fam in ['Kingdom', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', 'Genome', 'Genomes']:
                continue
            if fam_freq_df.iloc[ri][fam] > 0:
                group_fams[group].add(fam)

    # identify fams found in only one group
    unique_grp_fams = {}  # {grp: {fams}}
    for group in tqdm(group_fams, desc=f"Identifying {group_by} specific fams"):
        fams_in_grp = group_fams[group]
        other_groups = list(group_fams.keys())
        other_groups.remove(group)

        for fam in fams_in_grp:
            unique = True
            for grp in other_groups:
                if fam in group_fams[grp]:
                    unique = False

            if unique:
                try:
                    unique_grp_fams[group].add(fam)
                except KeyError:
                    unique_grp_fams[group] = {fam}

    return unique_grp_fams, group_fams



# /usr/bin/env python
# -*- coding: utf-8 -*-

# (c) University of St Andrews 2023
# (c) University of Strathclyde 2023
# (c) James Hutton Institute 2023
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""concatenate_cds.py

Script to concatenate CDS sequences and generate a single multigene
alignment file, with accompanying partition file so that model
parameters can be estimated for each gene.

The sequence ID is used to keep sequences from the same organism
together.
"""


import os

from pathlib import Path

from Bio import SeqIO
from tqdm import tqdm

from saintBioutils.utilities.file_io import make_output_directory

import argparse

from pathlib import Path
from typing import List, Optional


def main(argv: Optional[List[str]] = None):
    if argv is None:
        parser = build_parser()
        args = parser.parse_args()
    else:
        parser = build_parser(argv)
        args = parser.parse_args()

    make_output_directory(args.OUTDIR, force=args.force, nodelete=args.nodelete)

    # Directory containing genome sequences
    args.GENOMEDIR = Path("data/pecto_dic/tree/genomes")

    # Directory containing threaded CDS sequences for concatenation
    args.CDSDIR = Path("data/pecto_dic/tree/sco_cds_aligned")

    # Directory for concatenated multigene sequences
    args.OUTDIR = Path("data/pecto_dic/tree/concatenated_cds")

    # Contruct dictionaries relating genome ID to organism, and seqID to genomeID
    print("Parsing genome IDs and names")
    genomedict = {}
    for fpath in tqdm([_ for _ in args.GENOMEDIR.iterdir() if _.suffix == ".fna"]):
        genomeid = "_".join(fpath.stem.split("_")[:2])
        seq = next(SeqIO.parse(fpath.open("r"), "fasta"))
        organism = (
            " ".join(seq.description.split()[1:])
            .split("chromosome")[0]
            .split("_contig")[0]
            .split("contig")[0]
            .split("scaffold")[0]
            .split(",")[0]
            .split("Dickeya_sp")[0]
            .split("NODE")[0]
            .split("ID_1")[0]
        )
        genomedict[genomeid] = organism
    print(f"Parsed {len(genomedict)} genome names")

    print("Parsing sequence IDs for each genome")
    seqid_dict = {}
    for fpath in tqdm([_ for _ in args.GENOMEDIR.iterdir() if _.suffix == ".fna"]):
        genomeid = "_".join(fpath.stem.split("_")[:2])
        for seqid in [_.id for _ in SeqIO.parse(fpath.open("r"), "fasta")]:
            seqid_dict[seqid] = genomeid
    print(f"Parsed {len(seqid_dict)} sequence IDs")

    # Concatenate CDS sequences
    print("Concatenating CDS sequences")
    concatenated = {}
    partitions = []
    lastpos = 0
    for fpath in tqdm(args.CDSDIR.iterdir()):
        with fpath.open("r") as ifh:
            for cds in SeqIO.parse(ifh, "fasta"):
                seqid = "_".join(cds.id.split("_")[:-1])
                genomeid = seqid_dict[seqid]
                if genomeid not in concatenated:
                    cds.id = genomeid
                    concatenated[genomeid] = cds
                else:
                    concatenated[genomeid] += cds
            nextpos = lastpos + len(cds)
            partitions.append(f"GTR+FO+G4m+B, seqid={lastpos + 1}-{nextpos}")
            lastpos = nextpos
    print(f"Generated {len(concatenated)} concatenated sequences")

    # Write partition file
    os.makedirs(args.OUTDIR, exist_ok=True)
    with (args.OUTDIR / "concatenated.part").open("w") as ofh:
        ofh.write("\n".join(partitions) + "\n")

    # Write concatenated sequences
    with (args.OUTDIR / "concatenated.fasta").open("w") as ofh:
        seqlist = []
        for key, val in concatenated.items():
            val.id = key
            val.description = ""
            seqlist.append(val)
        SeqIO.write(seqlist, ofh, "fasta")


def build_parser(argv: Optional[List] = None):
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = argparse.ArgumentParser(
        prog="cw_query_database",
        description="Interrogate a local CAZyme database",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "GENOMEDIR",
        type=Path,
        help="Path to dir containing genomic sequences"
    )
    parser.add_argument(
        "CDSDIR",
        type=Path,
        help="Path to dir containing prodigal predicted CDS seqs"
    )
    parser.add_argument(
        "OUTDIR",
        type=Path,
        help="Path to dir to write output"
    )

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force writing to existing output dir",
    )

    parser.add_argument(
        "-n",
        "--nodelete",
        dest="nodelete",
        action="store_true",
        default=False,
        help="When called, content in the existing output dir is NOT deleted",
    )

    if argv is None:
        # parse command-line
        return parser
    else:
        # return namespace
        return parser.parse_args(argv)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Build CLI for invoke_dbcan.py"""


from asyncio import subprocess
import sys

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, Namespace
from typing import List, Optional

from pathlib import Path
from typing import List, Optional

from cazomevolve import __version__, __citation__
from cazomevolve.utilities.parsers import (
    get_fam_seqs_parser,
    run_blast_parser,
    run_diamond_parser,
    dl_acc_genomes_parser,
    download_genomes_parser,
    build_cazy_db_parser,
    get_cazy_parser,
    invoke_dbcan_parser,
    get_dbcan_parser,
    add_taxs_parser,
    explore_cazomes_parser,
)


def build_parser(argv: Optional[List] = None) -> Namespace:
    """Parse command-line arguments for script.

    :param argv: Namesapce, command-line arguments

    The script offers a single main, parser with subcommands for the actions:

    :Exploring CAZy family sequence diversity:
    - get_fam_seqs - downloads protein sequences for CAZymes in families of interest

    """
    # Create parser object
    parser_main = ArgumentParser(
        prog="cazomevolve",
        description="Explore CAZymes and CAZomes",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser_main.add_subparsers(
        title="subcommands", description="Valid subcommands",
    )

    # add args to the main parser
    parser_main.add_argument(
        "--version",
        action="store_true",
        default=False,
        help="Print version number"
    )
    parser_main.add_argument(
        "--citation",
        action="store_true",
        default=False,
        help="Print citation information"
    )
    parser_main.add_argument(
        "-l",
        "--log",
        dest="log",
        action="store",
        default=None,
        type=Path,
        help="logfile location",
    )
    parser_main.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="report verbose progress to log",
    )

    # add subcommand parser

    # Seq diversity subcommands
    get_fam_seqs_parser.build_parser(subparsers)
    run_blast_parser.build_parser(subparsers)
    run_diamond_parser.build_parser(subparsers)

    # download genomes
    dl_acc_genomes_parser.build_parser(subparsers)
    download_genomes_parser.build_parser(subparsers)

    # annotate cazome
    build_cazy_db_parser.build_parser(subparsers)
    get_cazy_parser.build_parser(subparsers)
    invoke_dbcan_parser.build_parser(subparsers)
    get_dbcan_parser.build_parser(subparsers)
    add_taxs_parser.build_parser(subparsers)

    # explroe cazomes
    explore_cazomes_parser.build_parser(subparsers)

    # Parse arguments
    # The list comprehension is to allow PosixPaths to be defined and passed in testing
    if argv is None:
        if len(sys.argv) == 1:
            argv = ["-h"]
        else:
            argv = sys.argv[1:]
    return parser_main.parse_args([str(_) for _ in argv])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scazome.explore.cazome_sizes.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import numpy as np

from argparse import Namespace

from cazomevolve.cazome.explore import pca


@pytest.fixture
def pca_objs(built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Genus', 'Species'])
    pca_obj, xscale = pca.perform_pca(built_fam_freq_df, len(built_fam_freq_df))
    return pca_obj, xscale


@pytest.fixture
def ie_fam_freq(built_fam_freq_df):
    cols = []
    i = 0
    for c in list(built_fam_freq_df.columns):
        if c in ['Genome', 'Genus', 'Species']:
            cols.append(c)
        else:
            if str(i).endswith('5'):
                cols.append(f"i_{c}")
            else:
                cols.append(f"e_{c}")
        i += 1

    built_fam_freq_df.columns = cols
    return built_fam_freq_df


def test_pca(built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Genus', 'Species'])
    pca.perform_pca(built_fam_freq_df, len(built_fam_freq_df))


def test_plot_explained_variance(pca_objs, built_fam_freq_df):
    g = pca.plot_explained_variance(
        pca_objs[0],
        len(built_fam_freq_df),
    );


def test_scree(pca_objs):
    pca.plot_scree(pca_objs[0])


def test_plot_pca(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
    )


def test_plot_pca_style(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
        hue_order=None,
        style='Genus',
        style_order=None,
    )


def test_plot_pca_style_order(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    orders = list(built_fam_freq_df['Genus'])
    orders.sort()
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
        hue_order=None,
        style='Genus',
        style_order=orders,
    )


def test_plot_pca_hue_order(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    orders = list(built_fam_freq_df['Genus'])
    orders.sort()
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
        hue_order=orders,
        style=None,
        style_order=None,
    )


def test_plot_pca_hue_order_style(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    orders = list(built_fam_freq_df['Genus'])
    orders.sort()
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
        hue_order=orders,
        style='Genus',
        style_order=None,
    )


def test_plot_pca_hue_order_style_order(pca_objs, built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    orders = list(built_fam_freq_df['Genus'])
    orders.sort()
    pca.plot_pca(
        pca_objs[0],
        pca_objs[1],
        built_fam_freq_df,
        1,
        2,
        'Genus',
        hue_order=orders,
        style='Genus',
        style_order=orders,
    )


def test_plot_loadings(pca_objs, built_fam_freq_df):
    pca.plot_loadings(
        pca_objs[0],
        built_fam_freq_df,
        1,
        2,
    )


def test_plot_loadings_styled(pca_objs, built_fam_freq_df):
    pca.plot_loadings(
        pca_objs[0],
        built_fam_freq_df,
        1,
        2,
        style=True,
    )


def test_plot_loadings_ie(pca_objs, ie_fam_freq):
    pca.plot_ie_loadings(
        pca_objs[0],
        ie_fam_freq,
        1,
        2,
    )


def test_plot_loadings_styled_ie(pca_objs, ie_fam_freq):
    pca.plot_ie_loadings(
        pca_objs[0],
        ie_fam_freq,
        1,
        2,
        style=True,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Retrieve CAZy canoncial CAZyme classifications"""


import logging
import re
import sys

from typing import List, Optional

from Bio import SeqIO
from cazy_webscraper.sql.sql_orm import get_db_connection, Session, Genbank, CazyFamily
from cazy_webscraper.sql.sql_interface.get_data.get_table_dicts import get_gbk_table_dict
from saintBioutils.utilities.file_io import make_output_directory
from saintBioutils.utilities.file_io.get_paths import get_file_paths
from tqdm import tqdm

from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    if str(args.output_dir.parent) != ".":
        make_output_directory(args.output_dir, args.force, args.nodelete)

    if str(args.fam_genome_list.parent) != ".":
        make_output_directory(args.fam_genome_list.parent, args.force, args.nodelete)

    if str(args.fam_genome_protein_list.parent) != ".":
        if str(args.fam_genome_list.parent) == str(args.fam_genome_protein_list.parent):
            make_output_directory(args.fam_genome_protein_list.parent, True, True)
        else:
            make_output_directory(args.fam_genome_protein_list.parent, args.force, args.nodelete)

    logger = logging.getLogger(__name__)

    # connect to the local CAZyme db
    connection = get_db_connection(args.database, args.sql_echo, False)

    # retrieve path to protein FASTA files
    fasta_files_paths = get_file_paths(args.input_dir, suffixes=['.fasta', '.faa'])

    if len(fasta_files_paths) == 0:
        logger.error(
            f"Found 0 fasta files in {args.input_dir}\n"
            "Check the path is correct. Terminating program"
        )
        sys.exit(1)

    logger.warning(f"Retrieved {len(fasta_files_paths)} FASTA files")

    logger.warning("Parsing CAZy db Genbanks table into dict")
    gbk_table_dict = get_gbk_table_dict(connection)
    logger.warning("Loading the GenBanks annotations into dict")

    for fasta_path in tqdm(fasta_files_paths, desc="Getting CAZy annotations"):
        get_cazy_annotations(fasta_path, gbk_table_dict, args, connection)

    closing_message('Get CAZy CAZymes', args)


def get_cazy_annotations(fasta_path, gbk_table_dict, args, connection):
    """Get the CAZy family annotations for each fasta file.

    Move empty fasta files to directory used as input for dbCAN.

    :param fasta_path: POSIX path to FASTA file
    :param gbk_table_dict: dict of data in the cazyme db Genbanks table
    :param args: cmd-line args parser
    :param connection: connection to a sqlite db

    Return nothing.
    """
    logger = logging.getLogger(__name__)

    # compile path to write out non-CAZy annotated proteins
    output_path = args.output_dir / fasta_path.name

    # extract genomic accession from the file name
    try:
        genomic_accession = re.findall(r"GCF_\d+\.\d{1,5}", fasta_path.name)[0]
    except IndexError:
        try:
            genomic_accession = re.findall(r"GCA_\d+\.\d{1,5}", fasta_path.name)[0]
        except IndexError:
            logger.warning(
                f"Could not retrieve genomic accession from\n{fasta_path}\n"
                "Skipping FASTA file"
            )
            return

    cazy_accessions = set(list(gbk_table_dict.keys()))  # gbk accessions in the local db

    # load sequences in proteome FASTA file into dict
    fasta_seqs = {}  # {protein acc: seq record}
    for record in SeqIO.parse(fasta_path, "fasta"):
        fasta_seqs[record.id] = record

    fasta_accessions = set(list(fasta_seqs.keys()))
    logger.warning(f"Loaded {len(fasta_accessions)} seq IDs from {fasta_path.name}")

    acc_in_cazy, acc_not_in_cazy = set(), set()  # ensure they are reset to 0

    acc_in_cazy = cazy_accessions & fasta_accessions
    logger.warning(f"Found {len(acc_in_cazy)} proteins in local CAZyme db")

    acc_not_in_cazy = fasta_accessions.difference(acc_in_cazy)
    logger.warning(f"{len(acc_not_in_cazy)} proteins not in the local CAZyme db")

    if len(acc_not_in_cazy) != 0:
        # gather seqs of prot not in cazy and write to a FASTA file
        not_in_cazy_seqs = []
        for acc in acc_not_in_cazy:
            not_in_cazy_seqs.append(fasta_seqs[acc])
        SeqIO.write(not_in_cazy_seqs, output_path, "fasta")

    if len(acc_in_cazy) != 0:
        # compile data to write the tab delimited lists
        fam_genome_data = ""
        fam_genome_protein_data = ""
        
        for prot_acc in tqdm(acc_in_cazy, desc="Retrieving CAZy family annotaions from the local CAZyme database"):
            with Session(bind=connection) as session:
                fam_query = session.query(Genbank, CazyFamily).\
                    join(CazyFamily, Genbank.families).\
                    filter(Genbank.genbank_accession == prot_acc).\
                    all()

                for result in fam_query:
                    fam = result[1].family.split("_")[0]  # make sure to remove subfamily classification
                    fam_genome_data += f"{fam}\t{genomic_accession}\n"
                    fam_genome_protein_data += f"{fam}\t{genomic_accession}\t{prot_acc}\n"

        # write out data
        with open(args.fam_genome_list, "a") as fh:
            fh.write(fam_genome_data)
        
        with open(args.fam_genome_protein_list, "a") as fh:
            fh.write(fam_genome_protein_data)

    return


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions to get data from the CAZy website (www.cazy.org)"""


from Bio import SeqIO

import time

import mechanicalsoup

from tqdm import tqdm
from requests.exceptions import ConnectionError, MissingSchema
from urllib3.exceptions import HTTPError, RequestError


def get_cazy_proteins(fasta_file):
    """Retrieving NCBI protein accessions from FASTA file of CAZy proteins

    :param fasta_file: Path to fasta file of CAZy fam protein seqs

    Return list of NCBI protein accessions
    """
    prot_accs = []

    for record in SeqIO.parse(fasta_file, "fasta"):
        prot_accs.append(record.id)

    return list(set(prot_accs))


def get_cazy_db_prots(cazy_family, characterised=False, structured=False):
    """Get the NCBI protein accessions for proteins in the structure or characterised tables
    from the CAZy website.
    
    :param cazy_family: str, name of CAZy family in CAZy format, e.g. GH1 not gh1
    :param characterised: bool, retrieved proteins listed as 'characterised' in CAZy
    :param structured: bool, retrieve proteins listed with structures in CAZy
    
    Return list of NCBI protein accessions or None if fails
    """
    urls = []  # [ [url, data type, col index for cazy website] ]
    if characterised:
        urls.append([f"http://www.cazy.org/{cazy_family}_characterized.html", 'characterised', 4])
    if structured:
        urls.append([f"http://www.cazy.org/{cazy_family}_structure.html", "structured", 3])

    all_proteins = []

    for url in urls:
        page, error_mss = get_page(
            url[0],
            max_tries=100
        )
        if page is None:
            print(f'Did not retrieve page for {cazy_family}: {url[1]}')
            print(error_mss)
            continue
    
        cazyme_table = page.select('table')[1]

        gbk_bs_elements = []

        for row in tqdm(cazyme_table.select("tr"), desc=f"Parsing {url[1]} table for {cazy_family}"):
            try:
                if (row.attrs["class"] == ['royaume']) and (row.text.strip() != 'Top'):
                    continue
            except KeyError:
                pass

            try:
                if (row.attrs["id"] == 'line_titre'):
                    continue
            except KeyError:
                pass

            try:
                gbk_bs_elements += [_ for _ in row.select("td")[url[2]].contents if getattr(_, "name", None) != "br"]
            except IndexError:
                pass

        ncbi_accessions = get_all_accessions(gbk_bs_elements)
    
        all_proteins += list(set(ncbi_accessions))
    
    return all_proteins


def get_all_accessions(bs_element_lst):
    """Retrieve all accessions listed in a cell from a CAZyme table.

    :param bs_element_list: list of BeautifulSoup element from cell in HTML table

    Return list of accessions."""
    accessions = []

    for bs_element in bs_element_lst:
        try:
            if bs_element.name == "a":  # Hyperlinked, extract accession and add to primary
                accessions.append(bs_element.text)
            elif bs_element.strip() != "":  # There is text in element
                accessions.append(bs_element)
        except TypeError:
            pass

    return accessions


def browser_decorator(func):
    """Decorator to re-invoke the wrapped function up to 10 times."""

    def wrapper(*args, **kwargs):
        tries, success, err = 0, False, None

        while not success and (tries < kwargs['max_tries']):
            try:
                response = func(*args, **kwargs)
            except (
                ConnectionError,
                HTTPError,
                OSError,
                MissingSchema,
                RequestError,
            ) as err_message:
                if (tries < kwargs['max_tries']):
                    print(
                        f"Failed to connect to CAZy on try {tries}/{kwargs['max_tries']}.\n"
                        f"Error: {err_message}"
                        "Retrying connection to CAZy in 10s"
                    )
                success = False
                response = None
                err = err_message
            if response is not None:  # response was successful
                success = True
            # if response from webpage was not successful
            tries += 1
            time.sleep(10)
            
        if (not success) or (response is None):
            print(f"Failed to connect to CAZy.\nError: {err}")
            return None, err
        else:
            return response, None

    return wrapper


@browser_decorator
def get_page(url, **kwargs):
    """Create browser and use browser to retrieve page for given URL.

    :param url: str, url to webpage
    :param args: cmd-line args parser
    :kwargs max_tries: max number of times connection to CAZy can be attempted

    Return browser response object (the page).
    """
    # create browser object
    browser = mechanicalsoup.Browser()
    # create response object
    page = browser.get(url, timeout=45)
    page = page.soup

    return page



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Add taxonomic classifications to tab delimited lists"""


import logging
import pandas as pd
import sys

from typing import List, Optional

from Bio import Entrez
from saintBioutils.utilities.logger import config_logger
from tqdm import tqdm

from cazomevolve.taxs.ncbi import add_ncbi_taxs
from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    if logger is None:
        config_logger(args)
    logger = logging.getLogger(__name__)

    Entrez.email = args.email

    if (args.FGP_FILE is None) and (args.FG_FILE is None):
        logger.warning(
            "No tab delimited files provided\nPlease provide at least one file:\n"
            "--FGP_FILE - tab delimited file with Fam Genome Protein \n"
            "--FG_FILE - tab delimited file with Fam Genome"
        )
        sys.exit(1)

    col_names = ['Genome']
    if args.kingdom:
        col_names.append('Kingdom')
    if args.phylum:
        col_names.append('Phylum')
    if args.tax_class:
        col_names.append('Class')
    if args.tax_order:
        col_names.append('Order')
    if args.tax_family:
        col_names.append('Family')
    if args.genus:
        col_names.append('Genus')
    if args.species:
        col_names.append('Species')

    logger.warning(f"Will be retrieving the following data:\n{col_names}")

    if len(col_names) == 1:
        logger.warning("Must specify at least one rank of lineage to be included")
        sys.exit(1)

    gtdb_df = load_gtdb_df(col_names, args)

    # gather tax info
    # genomes_tax_dict set of genomic acc to query ncbi with to get the latest tax classification
    # genomes_to_query, dict {genome: f"genome_{tax}_{tax}"}
    genomes_tax_dict, genomes_to_query = add_gtdb_taxs(gtdb_df, col_names, args)

    if len(genomes_to_query) > 0:
        logger.warning(f"Retrieving taxonomic lineages from NCBI for {len(genomes_to_query)} genomes")
        genomes_tax_dict = add_ncbi_taxs(genomes_tax_dict, genomes_to_query, col_names, args)
    
    if args.FGP_FILE is not None:
        write_tab_lists(args.FGP_FILE, genomes_tax_dict, col_names)

    if args.FG_FILE is not None:
        write_tab_lists(args.FG_FILE, genomes_tax_dict, col_names)

    # write out CSV file as well
    write_out_csv(genomes_tax_dict, col_names, args)

    closing_message('Add taxs', args)


def load_gtdb_df(col_names, args):
    """Loading in the GTDB database dump (TSV file) into a pandas dataframe.

    :param col_names: list of column names, genomes and all tax levels of interest
    :param args: CLI-args parser

    Return pandas dataframe
    """
    logger = logging.getLogger(__name__)

    if args.gtdb is None:
        logger.warning("No GTDB tsv file provided.\nRetrieving all classifications from NCBI")
        # build an empty dataframe with the desired column names
        gtdb_data = {}
        for col_name in col_names:
            gtdb_data[col_name] = []
        gtdb_df = pd.DataFrame(gtdb_data)
    
    else:
        gtdb_data = []
        dl_gtdb_df = pd.read_table(args.gtdb)
        dl_gtdb_df.columns = ['Genome', 'Tax']

        # separate output tax into genus and species
        for ri in tqdm(range(len(dl_gtdb_df)), desc="Parsing GTDB data"):
        # for ri in tqdm(range(len(dl_gtdb_df)), desc="Parsing GTDB data"):
            genome_taxonomy = [dl_gtdb_df.iloc[ri]['Genome']]
            genome_taxonomy = [_.replace("RS_","").replace("GB_","").strip() for _ in genome_taxonomy]

            tax_info = dl_gtdb_df.iloc[ri]['Tax'].split(";")
            for data in tax_info:
                if args.kingdom and (data.strip().startswith('d__')):
                    genome_taxonomy.append(data.replace('d__','').strip())

                elif args.phylum and (data.strip().startswith('p__')):
                    genome_taxonomy.append(data.replace('p__','').strip())

                elif args.tax_class and (data.strip().startswith('c__')):
                    genome_taxonomy.append(data.replace('c__','').strip())

                elif args.tax_order and (data.strip().startswith('o__')):
                    genome_taxonomy.append(data.replace('o__','').strip())

                elif args.tax_family and (data.strip().startswith('f__')):
                    genome_taxonomy.append(data.replace('f__','').strip())

                elif args.genus and (data.strip().startswith('g__')):
                    genome_taxonomy.append(data.replace('g__','').strip())

                elif args.species and (data.strip().startswith('s__')):
                    species = " ".join(data.strip().split(" ")[1:])  # remove genus from species name
                    genome_taxonomy.append(species)

            gtdb_data.append(genome_taxonomy)
            
        gtdb_df = pd.DataFrame(gtdb_data, columns=col_names)
    
    return gtdb_df


def add_gtdb_taxs(gtdb_df, col_names, args):
    """
    Build dict of genome: tax using GTDB data 
    AND identify genomes to query ncbi with to get the latest tax classification

    :param gtdb_df: pandas df with a genome col, and one col per tax level of interest
    :param col_names: list of col names, including Genomes and one col per tax level of interest
    :param args: CLI args parser

    Return
        :var genomes_tax_dict: dict {genome: f'{genome}_{tax}'}
        :var genomes_to_query: set, genomes accs to query ncbi with
    """
    logger = logging.getLogger(__name__)
    all_genomes = []

    if args.FGP_FILE is not None:
        df = pd.read_table(args.FGP_FILE, header=None)
        df.columns = ['Fam', 'Genome', 'Protein']
        all_genomes += list(df['Genome'])

    if args.FG_FILE is not None:
        df = pd.read_table(args.FG_FILE, header=None)
        df.columns = ['Fam', 'Genome']
        all_genomes += list(df['Genome'])

    all_genomes = set(all_genomes)

    genome_tax_dict = {}
    genomes_to_query = set()

    if len(gtdb_df) == 0:
        return {}, all_genomes  # query all genomes against NCBI

    for genome in tqdm(all_genomes, desc="Getting GTDB tax"):
        g_rows = gtdb_df[gtdb_df['Genome'] == genome]
        if len(g_rows) == 0:
            # try alternative acc
            if genome.startswith('GCA'):
                alt_genome = genome.replace('GCA_', 'GCF_')
            else:
                alt_genome = genome.replace('GCF_', 'GCA_')

            g_rows = gtdb_df[gtdb_df['Genome'] == alt_genome]
            if len(g_rows) == 0:
                # genome not in gtdb df
                logger.info(
                    f"GenBank and RefSeq version of accession {genome} was not in the GTDB database\n"
                    "Will retrieve taxonomic classifications from NCBI"
                )
                genomes_to_query.add(genome)
                continue
        
        tax = f"{genome}_"
        for col_name in col_names:
            tax_info = g_rows.iloc[0][col_name]
            tax += f"{tax_info}_"

        genome_tax_dict[genome] = tax[:-1]  # drop terminal '_' underscore

    return genome_tax_dict, genomes_to_query
        

def write_tab_lists(file_path, genomes_tax_dict, col_names):
    """Write out data to tad delimited lists

    :param file_path: path to tab delimited list
    :param genomes_tax_dict: dict {genome: f'{genome}_{tax}'}
    :param col_names: list, lineage ranks

    Return nothing
    """
    # FGP or FG
    logger = logging.getLogger(__name__)

    with open(file_path, 'r') as fh:
        data_lines = fh.read().splitlines()

    all_data = []
    for line in data_lines:
        new_data = line.split("\t")
        try:
            # fam = new_data[0]
            # genome_acc = new_data[1]
            # protein = new_data[2]
            new_data[1] = genomes_tax_dict[new_data[1]]  # get tax info for the genomic acc
        except KeyError:
            logger.warning(f"Could not retrieve tax data for {new_data[1]}")
            genome = f"{new_data[1]}_"
            for rank in col_names:
                genome += f"NaN_"
            new_data[1] = genome[:-1]  # replace the acc with f'{genome}_{tax}_{tax}'
        all_data.append(new_data)

    output_path = file_path.parent / f"{file_path.name}_taxs"

    logger.warning(f"Writing to {output_path}")

    with open(output_path, 'w') as fh:
        for line in all_data:
            data = '\t'.join(line)
            fh.write(f"{data}\n")


def write_out_csv(genomes_tax_dict, col_names, args):
    """Write out CSV file of taxonomy information.

    :param genomes_tax_dic: {genomes: f'genomes_tax'}
    :param col_names: list of lineage ranks
    :param args: cli args parser

    Return nothing
    """
    df_data = []
    for genome in tqdm(genomes_tax_dict, desc="Building tax df"):
        new_row = [genome]

        genome_tax_data = genomes_tax_dict[genome].split("_")
        tax_data = genome_tax_data[2:]

        for i in range((len(col_names)-1)):  # -1 because the first col name is 'Genomes'
            if (i+2) == len(col_names):  # last column name
                # take all items remaining in tax data
                data = " ".join(tax_data[i:])
                if len(data) != 0:
                    new_row.append(data)
            else:
                data = tax_data[i]
                if len(data) != 0:
                    new_row.append(data)
                    
        df_data.append(new_row)
        
    df = pd.DataFrame(df_data, columns=col_names)

    prefix = ""
    if args.outpath is None:
        if args.FGP_FILE is not None:
            parent_dir = args.FGP_FILE.parent
            prefix = "fgp_"
        if args.FG_FILE is not None:
            parent_dir = args.FG_FILE.parent
            prefix = "fg_"
        
        outpath = parent_dir / f"{prefix}genome_taxs.csv"

    else:
        outpath = args.outpath

    df.to_csv(outpath)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions to get data from the CAZy website (www.cazy.org)"""


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns


# define the colour palettes for annotating proteins
PALETTE = sns.color_palette(['#425df5', '#eb8913', '#19bfb4', '#db0d4e', '#15ab62', '#ffffff'])
PALETTE_DICT = {
    'cand': '#425df5',  # candidates
    'struct': '#eb8913',  # protein with structures in RCSB PDB
    'structCand': '#19bfb4',  # candidates with structures in RCSB PDB
    'func': '#db0d4e',  # candidates listed as 'characterised' in CAZy
    'funcCand': '#15ab62',  # proteins listed as 'characterised' in CAZy
    'nothing': '#ffffff',  # nothing to note about this protein
}


# Run this block before visualising the data

def plot_clustermap(
    df,
    fam,
    varaible,
    title=None,
    colour_scheme='rocket_r',
    fig_size=(25, 25),
    save_fig=None,
    dpi=100,
    annotate=False,
    char_only=False,
    candidates={}, structured_prots={}, characterised_prots={},
    palette_dict=PALETTE_DICT,
):
    """Plot a cluster map for the specified variable
    
    :param df: pandas dataframe
    :param fam: str, CAZy family of interest
    :param variable: df, name of column containing the variable to plot
    :param title: str, default none. Title of plot
    :param colour_scheme: str, default rocket_r, seaborn colour scheme for plot
    :param fig_size: tuple, len 2, default (25, 10)
    :param save_fig: str, path to save file, default none, don't save fig
    :param dpi: int, default 100, resolution of saved file image
    :param annotate: bool, add annotation of protein candidates, and functionally/structurally 
        characteirsed proteins
    :param char_only: bool, if set to true, only plot proteins labelled as candidates or
        functionally/structurally characteirsed proteins
    :param candidates: dict {fam: [prot accessions of proteins of interest]}
    :param structured_prots: dict {fam: [prot acc of proteins listed in the structure table in CAZy]}
    :param characterised_prots: dict {fam: [prot acc of proteins listed in the characterised table in CAZy]}
    
    Return seaborn plot
    """
    df = df[['qseqid', 'sseqid', varaible]]
    
    if char_only:  # plot only proteins that are candidates and functionally/structurally characteirsed proteins
        charactised_prots = characterised_prots[fam] + structured_prots[fam] + candidates[fam]
        df = df[df['qseqid'].isin(charactised_prots)]
        df = df[df['sseqid'].isin(charactised_prots)]
    
    heatmap_data = pd.pivot_table(df, index='qseqid', columns='sseqid', values=varaible)
    heatmap_data.columns = list(heatmap_data.columns)
    heatmap_data.index = list(heatmap_data.columns)
    heatmap_data = heatmap_data.fillna(0)
    
    if annotate:
        # add extra info on structural and functional characterisation of the family
        extra_data = []

        for prot in list(heatmap_data.columns):
            if prot in candidates[fam]:
                if prot in characterised_prots[fam]:
                    extra_data.append(palette_dict['funcCand'])
                elif prot in structured_prots[fam]:
                    extra_data.append(palette_dict['structCand'])
                else:
                    extra_data.append(palette_dict['cand'])

            elif prot in structured_prots[fam]:
                extra_data.append(palette_dict['struct'])

            elif prot in characterised_prots[fam]:
                extra_data.append(palette_dict['func'])

            else:
                extra_data.append(palette_dict['nothing'])

        fig = sns.clustermap(
            heatmap_data,
            cmap=colour_scheme,
            figsize=fig_size,
            row_colors=extra_data,
            col_colors=extra_data,
        );

        # extra data legend
        for label in list(palette_dict.keys()):
            fig.ax_row_dendrogram.bar(0, 0, color=palette_dict[label], label=label, linewidth=0)

        l3 = fig.ax_row_dendrogram.legend(title='Characterisation', loc='upper right', ncol=1)
    
    else:
        fig = sns.clustermap(
            heatmap_data,
            cmap=colour_scheme,
            figsize=fig_size,
        );
    
    if save_fig is not None:
        fig.savefig(save_fig, dpi=dpi);
    
    return fig


def plot_heatmap_of_clustermap(
    fig,
    df,
    fam,
    varaible,
    title=None,
    colour_scheme='rocket_r',
    fig_size=(25, 25),
    save_fig=None,
    dpi=100,
    annotate=False,
    char_only=False,
    candidates={}, structured_prots={}, characterised_prots={},
    palette_dict=PALETTE_DICT,
):
    """Generate a heatmap for the defined variable, with proteins plotted in the same order as the provided
    clustermap (fig)
    
    :param fig: seaborn clustergrid of entire family, default None, clustermap,
    :param df: pandas dataframe
    :param fam: str, CAZy family of interest
    :param variable: df, name of column containing the variable to plot
    :param title: str, default none. Title of plot
    :param colour_scheme: str, default rocket_r, seaborn colour scheme for plot
    :param fig_size: tuple, len 2, default (25, 10)
    :param save_fig: str, path to save file, default none, don't save fig
    :param dpi: int, default 100, resolution of saved file image
    :param annotate: bool, add annotation of protein candidates, and functionally/structurally 
        characteirsed proteins
    :param char_only: bool, if set to true, only plot proteins labelled as candidates or
        functionally/structurally characteirsed proteins
    :param candidates: dict {fam: [prot accessions of proteins of interest]}
    :param structured_prots: dict {fam: [prot acc of proteins listed in the structure table in CAZy]}
    :param characterised_prots: dict {fam: [prot acc of proteins listed in the characterised table in CAZy]}
    
    Return nothing
    """
    column_order = list(fig.__dict__['data2d'].keys())
    row_order = list(fig.__dict__['data2d'].index)
    
    df = df[['qseqid', 'sseqid', varaible]]
    
    if char_only:  # plot only proteins that are candidates and functionally/structurally characteirsed proteins
        charactised_prots = characterised_prots[fam] + structured_prots[fam] + candidates[fam]
        df = df[df['qseqid'].isin(charactised_prots)]
        df = df[df['sseqid'].isin(charactised_prots)]
    
    heatmap_data = pd.pivot_table(df, index='qseqid', columns='sseqid', values=varaible)
    heatmap_data.columns = list(heatmap_data.columns)
    heatmap_data.index = list(heatmap_data.columns)
    heatmap_data = heatmap_data.fillna(0)
    
    heatmap_data = heatmap_data.to_dict()  # {col: {row: value}}

    heatmap_df_data = {}

    for _prot in column_order:
        column_data = heatmap_data[_prot] # dict of {row: value} for the column
        
        for __prot in row_order:
            row_value = column_data[__prot]

            try:
                heatmap_df_data[_prot]  # column
            except KeyError:
                heatmap_df_data[_prot] = {}

            heatmap_df_data[_prot][__prot] = row_value
            
    if annotate:
        # add extra info on structural and functional characterisation of the family
        extra_data_col = []

        for prot in column_order:
            # candidate 1, funct candidate 0.75, structured 0.5, functional 0.25, nothing 0
            if prot in candidates[fam]:
                if prot in characterised_prots[fam]:
                    extra_data_col.append(palette_dict['funcCand'])
                elif prot in structured_prots[fam]:
                    extra_data_col.append(palette_dict['structCand'])
                else:
                    extra_data_col.append(palette_dict['cand'])

            elif prot in structured_prots[fam]:
                extra_data_col.append(palette_dict['struct'])

            elif prot in characterised_prots[fam]:
                extra_data_col.append(palette_dict['func'])

            else:
                extra_data_col.append(palette_dict['nothing'])

        extra_data_row = []

        for prot in row_order:
            # candidate 1, funct candidate 0.75, structured 0.5, functional 0.25, nothing 0
            if prot in candidates[fam]:
                if prot in characterised_prots[fam]:
                    extra_data_row.append(palette_dict['funcCand'])
                elif prot in structured_prots[fam]:
                    extra_data_row.append(palette_dict['structCand'])
                else:
                    extra_data_row.append(palette_dict['cand'])

            elif prot in structured_prots[fam]:
                extra_data_row.append(palette_dict['struct'])

            elif prot in characterised_prots[fam]:
                extra_data_row.append(palette_dict['func'])

            else:
                extra_data_row.append(palette_dict['nothing'])

        fig = sns.clustermap(
            heatmap_df_data,
            cmap=colour_scheme,
            figsize=fig_size,
            row_cluster=False,
            col_cluster=False,
            row_colors=extra_data_row,
            col_colors=extra_data_col,
        );            

        # extra data legend
        for label in list(palette_dict.keys()):
            fig.ax_row_dendrogram.bar(0, 0, color=palette_dict[label], label=label, linewidth=0)

        l3 = fig.ax_row_dendrogram.legend(title='Info', loc='upper right', ncol=1)
    
    else:
        fig = sns.clustermap(
            heatmap_df_data,
            cmap=colour_scheme,
            figsize=fig_size,
            row_cluster=False,
            col_cluster=False,
        );

    if save_fig is not None:
        fig.savefig(save_fig, dpi=dpi);
    
    fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests main parser

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest

from argparse import Namespace
from pathlib import Path

from cazomevolve.scripts import cazomevolve_script
from cazomevolve.utilities.parsers import (
    parse_cmd,
    get_fam_seqs_parser,
    run_blast_parser,
    run_diamond_parser,
    dl_acc_genomes_parser,
    download_genomes_parser,
    build_cazy_db_parser,
    get_cazy_parser,
    invoke_dbcan_parser,
    get_dbcan_parser,
    add_taxs_parser,
    explore_cazomes_parser,
)


def test_parser_get_fam_seqs(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['get_fam_seqs', 'email', 'cazy', 'families', 'out'])
    assert new_namespace.email == 'email'
    assert new_namespace.cazy == 'cazy'
    assert new_namespace.families == 'families'
    assert new_namespace.force == False
    assert new_namespace.log == None
    assert new_namespace.citation == False


def test_run_blast_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['run_fam_blast', 'fasta', 'outfile'])
    assert new_namespace.fasta == 'fasta'
    assert new_namespace.outfile == 'outfile'


def test_run_diamond_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['run_fam_diamond', 'fasta', 'diamond', 'outfile'])
    assert new_namespace.fasta == 'fasta'
    assert new_namespace.outfile == 'outfile'
    assert new_namespace.diamond_db == 'diamond'


def test_dl_acc_genomes_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['download_acc_genomes', 'acc', 'outdir', 'genbank', 'genbank'])
    assert new_namespace.accessions == 'acc'
    assert new_namespace.outdir == 'outdir'
    assert new_namespace.file_opts == 'genbank'


def test_dl_acc_genomes_parser_invalid_format(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_acc_genomes', 'acc', 'outdir', 'ggggg', 'genbank'])
    assert pytest_wrapped_e.type == SystemExit
    

def test_dl_acc_genomes_parser_invalid_db(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_acc_genomes', 'acc', 'outdir', 'fasta', 'database'])
    assert pytest_wrapped_e.type == SystemExit


def test_dl_acc_genomes_parser_invalid_level(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_acc_genomes', 'acc', 'outdir', 'fasta', 'genbank', 'levels'])
    assert pytest_wrapped_e.type == SystemExit


def test_download_genomes_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['download_genomes', 'email', 'out', 'pectobacteriaceae', 'genomic', 'genbank'])
    assert new_namespace.email == 'email'
    assert new_namespace.terms == 'pectobacteriaceae'
    assert new_namespace.database == 'genbank'


def test_dl_genomes_parser_invalid_format(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_genomes', 'email', 'out', 'Aspergillus', 'ERROR', 'genbank'])
    assert pytest_wrapped_e.type == SystemExit
    

def test_dl_genomes_parser_invalid_db(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_genomes', 'email', 'out', 'Aspergillus', 'genomic', 'ERROR'])
    assert pytest_wrapped_e.type == SystemExit


def test_dl_genomes_parser_invalid_level(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['download_genomes', 'email', 'out', 'Aspergillus', 'genomic', 'genbank', 'ERROR'])
    assert pytest_wrapped_e.type == SystemExit


def test_build_cazy_db_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['build_cazy_db', 'email', 'db'])
    assert new_namespace.email == 'email'


def test_get_cazy_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['get_cazy_cazymes', 'in', 'db', 'out', 'fam_genome_list', 'fam_genome_protein_list'])
    assert new_namespace.database == Path('db')


def test_invoke_dbcan_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['run_dbcan', 'in', 'out', '3'])
    assert new_namespace.dbcan_version == 3


def test_get_dbcan_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['get_dbcan_cazymes', 'in', 'fam_genome', 'fam_genome_protein'])
    assert new_namespace.fam_genome_protein_list == Path('fam_genome_protein')


def test_add_taxs_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['add_taxs', 'email'])
    assert new_namespace.email == 'email'


def test_explore_cazomes_parser(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    new_namespace = parse_cmd.build_parser(['explore_cazomes', 'fgp_file', 'tax_csv_path', 'out'])
    assert new_namespace.fgp_file == Path('fgp_file')
    assert new_namespace.tax_csv_path == Path('tax_csv_path')


def test_explore_cazomes_parser_invalid_foramt(monkeypatch):
    def mock_run(*args, **kwards):
        return

    monkeypatch.setattr(cazomevolve_script, "main", mock_run)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        new_namespace = parse_cmd.build_parser(['explore_cazomes', 'fgp_file', 'tax_csv_path', 'out', 'prot', 'genus', 'ppp'])
    assert pytest_wrapped_e.type == SystemExit



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Build args parser for get_cazy_cazymes.py"""


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, _SubParsersAction, Action
from pathlib import Path
from typing import List, Optional

from cazomevolve.cazome.explore import explore_cazomes


class ValidateFormats(Action):
    """Check the user has provided valid database options."""
    def __call__(self, parser, args, values, option_string=None):
        valid_formats = ['png', 'pdf', 'svg']
        invalid = False
        if values[0] not in valid_formats:
            invalid = True
            raise ValueError(f'Invalid file format "{values[0]}" provided. Accepted formats: {valid_formats}')
        if invalid:
            sys.exit(1)
        setattr(args, self.dest, values[0])


def build_parser(
    subps: _SubParsersAction, parents: Optional[List[ArgumentParser]] = None
) -> None:
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = subps.add_parser(
        "explore_cazomes", formatter_class=ArgumentDefaultsHelpFormatter
    )

    # Add positional arguments to parser

    parser.add_argument(
        "fgp_file",
        type=Path,
        help="Path to tab delimited FGP file, listing CAZy families, genomic accessions, and protein IDs",
    )

    parser.add_argument(
        "tax_csv_path",
        type=Path,
        default=None,
        help="Path CSV file listing taxonomic data, containing a column called 'Genome', listing genomic accessions, and one column per taxonomic rank retrieved from NCBI and/or GTDB.",
    )

    parser.add_argument(
        "output_dir",
        type=Path,
        help="Directory to write out all outputs",
    )

    parser.add_argument(
        "--proteome_dir",
        type=Path,
        default=None,
        help=(
            "Path to directory containing protein FASTA files -\n"
            "Use this flag to calculate the percentage of the proteome that encapsulates the CAZome"
        ),
    )

    parser.add_argument(
        "--group_by",
        type=str,
        default='Genus',
        help="Taxonomic rank to group data by. Default: Genus. Will calculate means and SDs for these groups",
    )

    parser.add_argument(
        "--formats",
        metavar='Figure file formats',
        action=ValidateFormats,
        choices=['png', 'pdf', 'svg'],
        nargs='+',
        type=str,
        default='pdf',
        help="File formats to write out figures. Can specify multiple foramts. Default write out all figures in PDF foramt",
    )

    parser.add_argument(
        "--kingdom",
        dest="kingdom",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains Kingdom lineage",
    )

    parser.add_argument(
        "--phylum",
        dest="phylum",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains phylum lineage",
    )

    parser.add_argument(
        "--tax_class",
        dest="tax_class",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains class lineage",
    )

    parser.add_argument(
        "--tax_order",
        dest="tax_order",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains order lineage",
    )

    parser.add_argument(
        "--tax_family",
        dest="tax_family",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains (tax) family lineage",
    )

    parser.add_argument(
        "--genus",
        dest="genus",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains genus lineage",
    )

    parser.add_argument(
        "--species",
        dest="species",
        action="store_true",
        default=False,
        help=" Taxonomy CSV file contains species lineage",
    )

    parser.add_argument(
        "--show_plots",
        dest="show_plots",
        action="store_true",
        default=False,
        help="Display generated plots during run",
    )

    parser.add_argument(
        "--round_by",
        type=int,
        default=2,
        help="Number of decimal places to round means and SDs to",
    )

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force file over writting",
    )
    
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        metavar="log file name",
        default=None,
        help="Defines log file name and/or path",
    )
    
    parser.add_argument(
        "-n",
        "--nodelete",
        dest="nodelete",
        action="store_true",
        default=False,
        help="enable/disable deletion of exisiting files",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Set logger level to 'INFO'",
    )
    parser.set_defaults(func=explore_cazomes.main)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions to parse diamond/blast data"""


import pandas as pd

from tqdm import tqdm


# Run this block before loading data

def load_data(data_file, fam):
    """Load the output data from BLASTP+/diamond into a pandas dataframe, and calculate the BLAST score ratio.
    
    :param data_file: str, path to the DIAMOND output file
    :param fam: str, name of CAZy family
    
    Return a pandas dataframe
    """
    df = pd.read_csv(data_file, sep='\t', header=None)
    
    column_names = ['qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore']
    df.columns = column_names
    
    df['BSR'] = df['bitscore'] / df['qlen']
    
    df['qcov'] = df['length'] / df['qlen']
    df['scov'] = df['length'] / df['slen']
    
    df = remove_redunant_prots(df, fam)
    
    return df


def remove_redunant_prots(df, fam, candidates={}, structured_prots={}, characterised_prots={}):
    """Identify groups of identical proteins, and retaining only one member per group

    If providing candidates, structure_prots, and/or characterised all proteins in these dicts 
    will be kept even if protein seqs are identical.
    
    :param df: pandas df containing data from DIAMOND
    :param fam: str, name of CAZy family - group name in the CANDIDATES df

    :param candidates: dict {fam: [prot accessions of proteins of interest]}
    :param structured_prots: dict {fam: [prot acc of proteins listed in the structure table in CAZy]}
    :param characterised_prots: dict {fam: [prot acc of proteins listed in the characterised table in CAZy]}
    
    return df
    """

    # redundant proteins have a qcov of 1 and a pi of 100
    redundance_df = df.loc[ ( (df['pident'] == 100) & (df['qcov'] == 1) ) ]
    # find  the groups of redundant proteins
    redundant_grps = {}

    grp_num = 0

    parsed_prots = set()

    # identify groups of redundant proteins
    for ri in tqdm(range(len(redundance_df)), desc="Identifying IPGs"):
        row = redundance_df.iloc[ri]

        qseqid = row['qseqid']

        if qseqid in parsed_prots:
            continue  # has already been added

        # get all rows with the same query seq id
        qseqid_rows = redundance_df.loc[redundance_df['qseqid'] == qseqid]

        if len(qseqid_rows) == 1:
            continue  # aligned against self only

        subject_ids_to_add = set()

        # for each subject id
        # check if the versus is true, the qseqid is the sseqid when the sseqid is the qseqid
        for q_ri in range(len(qseqid_rows)):
            q_row = qseqid_rows.iloc[q_ri]
            sub_seqid = q_row['sseqid']

            # retrieve the row where the qseqid is now the subject, and the subject id is now the query seq
            # They are already in the redundancy df, therefore pident is 100 and qcov is 1
            sseqid_rows = redundance_df.loc[(
                (redundance_df['qseqid'] == sub_seqid) &
                (redundance_df['sseqid'] == qseqid))
            ]

            if len(sseqid_rows) > 0:
                subject_ids_to_add.add(sub_seqid)

        if len(subject_ids_to_add) > 0:
            # found redunant pairs for qseqid
            redundant_grps[grp_num] = {qseqid}

            for sub_seqid in subject_ids_to_add:
                redundant_grps[grp_num].add(sub_seqid)
                parsed_prots.add(sub_seqid)

            grp_num += 1

        parsed_prots.add(qseqid)

    # from each group select a representative protein
    # and identify members of the group that will be dropped
    removing = set()
    
    print(f"Identified {len(list(redundant_grps.keys()))} groups of identical proteins")

    for grp in redundant_grps:
        prots_to_keep = set()
        
        for prot in redundant_grps[grp]:
            try:
                # retain proteins marked as candidates, functionally characitersed or structurally characterised
                if prot in candidates[fam]:
                    prots_to_keep.add(prot)
                elif prot in structured_prots[fam]:
                    prots_to_keep.add(prot)
                elif prot in characterised_prots[fam]:
                    prots_to_keep.add(prot)
                elif len(prots_to_keep) == 0: # ensure at least one protein from the group is retained
                    prots_to_keep.add(prot)
                else:  # already have members from the group so drop the protein
                    removing.add(prot)
            except KeyError:
                if len(prots_to_keep) == 0:
                    prots_to_keep.add(prot)
                else:  # already have members from the group so drop the protein
                    removing.add(prot)

    df = df[~df['qseqid'].isin(removing)]
    df = df[~df['sseqid'].isin(removing)]
    
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scripts in the script module

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import subprocess

from argparse import Namespace

from saintBioutils.utilities import logger, file_io

from cazomevolve.scripts import (
    build_cazy_db,
    cazomevolve_script,
    download_acc_genomes,
    get_fam_seqs,
    run_fam_blast,
    run_fam_diamond
)
from cazomevolve.utilities.parsers import parse_cmd


def test_main_citation(monkeypatch):
    """Test the main entry point for cazomevolve"""
    def mock_build_parser(*args, **kwards):
        return Namespace(citation=True, version=False, cmdline=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])
    
    def mock_build_logger(*args, **kwards):
        return
    
    monkeypatch.setattr(parse_cmd, "build_parser", mock_build_parser)
    monkeypatch.setattr(logger, "config_logger", mock_build_logger)
    monkeypatch.setattr(cazomevolve_script, "config_logger", mock_build_logger)

    cazomevolve_script.main(argv=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])


def test_main_version(monkeypatch):
    """Test the main entry point for cazomevolve"""
    def mock_build_parser(*args, **kwards):
        return Namespace(citation=False, version=True, cmdline=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])
    
    def mock_build_logger(*args, **kwards):
        return
    
    monkeypatch.setattr(parse_cmd, "build_parser", mock_build_parser)
    monkeypatch.setattr(logger, "config_logger", mock_build_logger)
    monkeypatch.setattr(cazomevolve_script, "config_logger", mock_build_logger)

    cazomevolve_script.main(argv=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])


def test_main(monkeypatch):
    """Test the main entry point for cazomevolve"""
    def mock_build_parser(*args, **kwards):
        return Namespace(citation=False, version=True, cmdline=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])
    
    def mock_build_logger(*args, **kwards):
        return
    
    def mock_run(*args, **kwards):
        return
    
    monkeypatch.setattr(parse_cmd, "build_parser", mock_build_parser)
    monkeypatch.setattr(logger, "config_logger", mock_build_logger)
    monkeypatch.setattr(cazomevolve_script, "config_logger", mock_build_logger)
    monkeypatch.setattr(get_fam_seqs, "main", mock_run)

    cazomevolve_script.main(argv=['get_fam_seqs', 'email', 'cazy', 'families', 'out'])


def test_build_cazy_df(monkeypatch):
    """Test the main entry point for build_cazy_db.py"""    
    def mock_run(*args, **kwards):
        return
    
    monkeypatch.setattr(subprocess, "Popen", mock_run)

    args = Namespace(email='email', db='db')

    build_cazy_db.main(args)


def test_download_acc_genomes(monkeypatch):
    """Test the main entry point for download_acc_genomes.py"""    
    def mock_run_1(*args, **kwards):
        return
    
    monkeypatch.setattr(subprocess, "call", mock_run_1)

    args = Namespace(
        accessions='acc',
        outdir='outdir',
        file_opts='fasta',
        database='database',
        assembly_levels='all',
    )

    download_acc_genomes.main(args)


def test_get_fam_seqs(monkeypatch):
    """Test the main entry point for get_fam_seqs.py"""    
    def mock_run_1(*args, **kwards):
        return
    
    monkeypatch.setattr(subprocess, "run", mock_run_1)
    monkeypatch.setattr(get_fam_seqs, "make_output_directory", mock_run_1)
    monkeypatch.setattr(file_io, "make_output_directory", mock_run_1)

    args = Namespace(
        cazy='cazy',
        outdir='outdir',
        email='email',
        families='GH1,GH2',
        force=False,
        nodelete=False,
    )

    get_fam_seqs.main(args)


def test_run_fam_blast(monkeypatch):
    """Test the main entry point for run_fam_blast.py"""    
    def mock_run_1(*args, **kwards):
        return
    
    monkeypatch.setattr(subprocess, "call", mock_run_1)

    args = Namespace(
        fasta='fasta',
        outfile='outfile',
    )

    run_fam_blast.main(args)


def test_run_fam_diamond(monkeypatch):
    """Test the main entry point for run_fam_diamond.py"""    
    def mock_run_1(*args, **kwards):
        return
    
    monkeypatch.setattr(subprocess, "call", mock_run_1)

    args = Namespace(
        fasta='fasta',
        diamond_db='diamond_db',
        outfile='outfile',
    )

    run_fam_diamond.main(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions for running and plotting PCA (principal component analysis)"""


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from tqdm import tqdm
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

import adjustText


def perform_pca(df, nComp):
    """Perform PCA on family freq df
    
    :param df: df, rows=genomes, cols=fam freqs
        Only contains columns with CAZy family frequency data
        Recommend placing the tax data in the index or leaving out
    :param nComp: int, number of components
     
    Return PCA object and object for scaling PCA
    """
    # scale the data
    scaler = StandardScaler()
    scaler.fit(df.loc[:, df.columns])
    X_scaled = scaler.transform(df.loc[:, df.columns])

    cazome_pca = PCA(n_components=nComp)
    cazome_pca.fit(X_scaled)
    
    return cazome_pca, X_scaled


def plot_explained_variance(
    pca,
    nComp,
    threshold=0.95,
    figsize=(10, 5),
    file_path=None,
    file_format='png',
    dpi=300,
    show=False,
):
    """Plot the cumlative explained variance.
    
    :param pca: sklearn pca object
    :param nComp: int, number of PCs
    :param threshold: int (float), calc the number of PCs needed to capture this degree of the variance 
        in the dataset
    :param file_path: str/Path, path to write out figure. If none, image is not written to a file
    :param file_format: str, format to write out the image. Default, png
    :param dpi: int, resolution for saved figure
    
    Retrun numpy array with the explained variance per PC
    """
    plt.clf()
    cumExpVar = np.cumsum(pca.explained_variance_ratio_)

    keepPC = [pc for pc in range(nComp) if cumExpVar[pc] >= threshold][0]
    print(
        'Number of features needed to explain {:1.2f} fraction of total variance is {:2d}. '.format(threshold, keepPC)
    )

    fig = plt.figure(figsize=figsize)
    im = plt.plot( range(nComp), cumExpVar )
    plt.xticks(np.arange(0,nComp,5));
    plt.xlabel( 'Number of PCs', fontsize=14);
    plt.ylabel('Cumulative explained variance', fontsize=14);
    if show:
        plt.show()

    if file_path is not None:
        plt.savefig(
            file_path,
            bbox_inches='tight',
            format=file_format,
            dpi=dpi,
        )

    return cumExpVar
    
    
def plot_scree(pca, nComp=10, file_path=None, file_format='png', dpi=300, show=False):
    """Generate scree plot for PCA, plotting the amount of variance captured by each pc, for the
    first nComp PCs
    
    :param pca: sklearn pca object
    :param nComp: int, number of PCs to plot
    :param file_path: str/Path, path to write out figure. If none, image is not written to a file
    :param file_format: str, format to write out the image. Default, png
    :param dpi: int, resolution for saved figure
    
    Return nothing
    """
    plt.clf()
    
    PC_values = np.arange(nComp) + 1
    plt.plot(PC_values, pca.explained_variance_ratio_[0:nComp], 'o-', linewidth=2, color='blue')
    plt.xlabel('Principal Component')
    plt.ylabel('Variance Explained')
    if file_path is not None:
        plt.savefig(
            file_path,
            bbox_inches='tight',
            dpi=dpi,
            format=file_format,
        )
    if show:
        plt.show();

    for i in range(nComp):
        print(f"Explained variance for {i+1}PC: {pca.explained_variance_ratio_[i]}")
    

def plot_pca(
    pca,
    X_scaled,
    fam_df,
    first_pc,
    second_pc,
    group_by,
    file_path=None,
    file_format='png',
    style=None,
    style_order=None,
    hue_order=None,
    font_scale=1.15,
    figsize=None, 
    xlim=None,
    ylim=None,
    dpi=300,
    loc='upper left',
    marker_size=100,
    markers=True,
    ax=None,
    show=False,
):
    """Project genomes onto the PCs
    
    :param pca: sklearn PCA object
    :param X_scaled: obj from scaling data
    :param fam_df: df of cazy family freqs
    :param first_pc: int, number of the first PC
    :param second_pc: int, number of the second PC
    :param group_by: how to group/colour data, genus or species
    
    OPTIONS
    :param file_path: path to write out fig, if none no file saved
    :param style: str, name of column to use to define style/marker style
    :param style_order: list order to list styles
    :param hue_order: list to write/assign categories of colours
    :param font_scale: float, scale font. >1 increases font size
    :param xlim: tuple, limits of the x axis
    :param ylim: tuple, limits of the y axis
    :param dpi: int, dpi to write out figure
    :param loc: str, location of key
    :param marker_size: int, scale of markers
    :param markers: dict, pass dict to map each level style to a marker 
        defined in matplotlib
    :param ax: axis - used to combining multiple scatter plots into a multiplot
    
    Return plot
    """
    grouping = f"{group_by[0].upper()}{group_by[1:]}"
    X_pca = pca.transform(X_scaled)
    
    if figsize is not None:
        plt.figure(figsize=figsize)
        
    sns.set(font_scale=font_scale)

    if hue_order is not None:
        print('Applying hue order')
        
        if style is not None:
            print('Applying style')
            
            if style_order is not None:
                print('Applying style order')
                # all options specified
                # apply style order
                g = sns.scatterplot(
                    x=X_pca[:,first_pc-1],
                    y=X_pca[:, second_pc-1],
                    data=fam_df,
                    hue=group_by,
                    s=marker_size,
                    hue_order=hue_order,
                    style=style,
                    style_order=style_order,
                    markers=markers,
                    ax=ax,
                )
                
            else:
                print('Not applying style order')
                # use default style order
                g = sns.scatterplot(
                    x=X_pca[:,first_pc-1],
                    y=X_pca[:, second_pc-1],
                    data=fam_df,
                    hue=group_by,
                    s=marker_size,
                    hue_order=hue_order,
                    style=style,
                    markers=markers,
                    ax=ax,
                )                
            
        else:
            print('Not applying style')
            # hue order only
            g = sns.scatterplot(
                x=X_pca[:,first_pc-1],
                y=X_pca[:, second_pc-1],
                data=fam_df,
                hue=group_by,
                s=marker_size,
                hue_order=hue_order,
                markers=markers,
                ax=ax,
            )  
        
    else:  # using default hue order - i.e. order data is presented in df
        print('Not applying hue order')
        
        if style is not None:  # use different markers for catagroies in provided col
            print('Applying style')
            
            if style_order is not None:  # define the order of the marker styles
                print('Applying style order')
                # apply style order
                g = sns.scatterplot(
                    x=X_pca[:,first_pc-1],
                    y=X_pca[:, second_pc-1],
                    data=fam_df,
                    hue=group_by,
                    s=marker_size,
                    style=style,
                    style_order=style_order,
                    markers=markers,
                    ax=ax,
                )
                
            else:
                print('Not applying style order')
                # use default style order
                g = sns.scatterplot(
                    x=X_pca[:,first_pc-1],
                    y=X_pca[:, second_pc-1],
                    data=fam_df,
                    hue=group_by,
                    s=marker_size,
                    style=style,
                    markers=markers,
                    ax=ax,
                )
        
        else:
            print('Not Applying style')
            # no options specified
            # do not apply style
            g = sns.scatterplot(
                x=X_pca[:,first_pc-1],
                y=X_pca[:, second_pc-1],
                data=fam_df,
                hue=group_by,
                s=marker_size,
                markers=markers,
                ax=ax,
            )
    
    if xlim is not None:
        g.set(xlim=xlim);
    if ylim is not None:
        g.set(ylim=ylim);
    
    g.axhline(0, linestyle='--', color='grey', linewidth=1.25);
    g.axvline(0, linestyle='--', color='grey', linewidth=1.25);
    
    plt.ylabel(f"PC{second_pc} {100 * pca.explained_variance_ratio_[(second_pc - 1)]:.2f}%");
    plt.xlabel(f"PC{first_pc} {100 * pca.explained_variance_ratio_[(first_pc - 1)]:.2f}%");
    plt.legend(bbox_to_anchor=(1.02, 1), loc=loc, borderaxespad=0);
    sns.move_legend(g, "lower center", bbox_to_anchor=(.5, 1), ncol=3, title=None, frameon=False);
    
    if file_path is not None:
        plt.savefig(
            file_path,
            bbox_inches='tight',
            dpi=dpi,
            format=file_format,
        )
    if show:
        plt.show();
    
    return plt


def plot_loadings(
    pca,
    fam_df,
    first_pc,
    second_pc,
    style=False,
    threshold=0.7,
    font_scale=1.15,
    font_size=12,
    dpi=300,
    fig_size=(16,16),
    file_path=None,
    file_format='png',
    marker_size=100,
    ax=None,
    show=False,
):
    """Build loadings plot
    
    :param pca: sklearn pca object
    :param fam_df: cazy family frequncy df
    :param first_pc: int, number of the first PC, e.g. PC1 == 1
    :param second_pc: int, number of the second PC e.g. PC2 == 2
    :param style: boolean, change shape of points depending on CAZy class
    :param threshold: correlation cut off for showing labels
        Only families with a value greater than the threshold
        will be annotated
    :param font_scale: scale font
    :param font_size: font size of family labels
    :param fig_size: tuple (width, height) of final plot
    :param file_path: str, path to write out a figure.
        If None, no figure is saved
    :param ax: axis, used to define axis in multiple plot to place figure
    
    Return nothing"""
    sns.set(font_scale=font_scale)

    # calculate loading = variables x loadings, returns an array
    loadings = pca.components_.T * np.sqrt(pca.explained_variance_)
    # get labels of variables, i.e. cazy families
    loadings_labels = list(fam_df.columns)

    for col in ['Kingdom', 'Phylum', 'Class', 'Order', 'Tax_Family', 'Genus', 'Species', 'Genome', 'Genomes']:
        try:
            loadings_labels.remove(col)
        except (KeyError, ValueError):
            pass

    loadings_x = loadings[:,(first_pc-1)]
    loadings_y = loadings[:,(second_pc-1)]

    loadings_df = pd.DataFrame()
    loadings_df['loadings_x'] = loadings_x
    loadings_df['loadings_y'] = loadings_y

    cazy_class = []
    for lbl in loadings_labels:
        if lbl.startswith('GH'):
            cazy_class.append('GH')
        elif lbl.startswith('GT'):
            cazy_class.append('GT')
        elif lbl.startswith('PL'):
            cazy_class.append('PL')
        elif lbl.startswith('CE'):
            cazy_class.append('CE')
        elif lbl.startswith('AA'):
            cazy_class.append('AA')
        else:
            cazy_class.append('CBM')

    loadings_df['cazy_class'] = cazy_class

    plt.figure(figsize=fig_size)
    if style:
        g = sns.scatterplot(
            x=loadings_x,
            y=loadings_y,
            data=loadings_df,
            hue=cazy_class,
            s=marker_size,
            style=cazy_class,
            ax=ax,
        );
    else:
        g = sns.scatterplot(
            x=loadings_x,
            y=loadings_y,
            data=loadings_df,
            hue=cazy_class,
            s=marker_size,
            ax=ax,
        );
    g.axhline(0, linestyle='--', color='grey', linewidth=1.25);
    g.axvline(0, linestyle='--', color='grey', linewidth=1.25);
    g.set(xlim=(-1,1),ylim=(-1,1));
    plt.ylabel(f"PC{second_pc}") 
    plt.xlabel(f"PC{first_pc}")

    texts = [
        plt.text(
            xval,
            yval,
            lbl,
            ha='center',
            va='center',
            fontsize=font_size,
        ) for (xval, yval, lbl) in zip(
            loadings_x, loadings_y, loadings_labels
        ) if abs(xval) > threshold or abs(yval) > threshold
    ]
    adjustText.adjust_text(texts, arrowprops=dict(arrowstyle='-', color='black'));

    sns.move_legend(g, "lower center", bbox_to_anchor=(.5, 1), ncol=3, title=None, frameon=False);
    
    if show:
        plt.show();

    if file_path is not None:
        plt.savefig(file_path, dpi=dpi, bbox_inches='tight', format=file_format)


def plot_ie_loadings(
    pca,
    fam_df,
    first_pc,
    second_pc,
    style=False,
    threshold=0.7,
    font_scale=1.15,
    font_size=12,
    dpi=300,
    fig_size=(16,16),
    file_path=None,
    marker_size=100,
):
    """Build loadings plot
    
    Modified from cazomevolve - styles points using intracellular/extracellular classification
    
    :param pca: sklearn pca object
    :param fam_df: cazy family frequncy df
    :param first_pc: int, number of the first PC, e.g. PC1 == 1
    :param second_pc: int, number of the second PC e.g. PC2 == 2
    :param threshold: correlation cut off for showing labels
        Only families with a value greater than the threshold
        will be annotated
    :param font_scale: scale font
    :param font_size: font size of family labels
    :param fig_size: tuple (width, height) of final plot
    :param file_path: str, path to write out a figure.
        If None, no figure is saved
    
    Return nothing"""
    sns.set(font_scale=font_scale)

    # calculate loading = variables x loadings, returns an array
    loadings = pca.components_.T * np.sqrt(pca.explained_variance_)
    # get labels of variables, i.e. cazy families
    loadings_labels = list(fam_df.columns)
    for col in ['Kingdom', 'Phylum', 'Class', 'Order', 'Tax_Family', 'Genus', 'Species', 'Genome', 'Genomes']:
        try:
            loadings_labels.remove(col)
        except (KeyError, ValueError):
            pass
    loadings_x = loadings[:,(first_pc-1)]
    loadings_y = loadings[:,(second_pc-1)]

    loadings_df = pd.DataFrame()
    loadings_df['loadings_x'] = loadings_x
    loadings_df['loadings_y'] = loadings_y

    cazy_class = []
    for lbl in loadings_labels:
        if lbl.find('GH') != -1:
            cazy_class.append('GH')
        elif lbl.find('GT') != -1:
            cazy_class.append('GT')
        elif lbl.find('PL') != -1:
            cazy_class.append('PL')
        elif lbl.find('CE') != -1:
            cazy_class.append('CE')
        elif lbl.find('AA') != -1:
            cazy_class.append('AA')
        else:
            cazy_class.append('CBM')

    loadings_df['cazy_class'] = cazy_class
    
    ie_classifications = []
    for lbl in loadings_labels:
        if lbl.startswith("i_"):
            ie_classifications.append('Intracellular')
        else:
            ie_classifications.append('Extracellular')
    loadings_df['ie_classification'] = ie_classifications

    plt.figure(figsize=fig_size)
    g = sns.scatterplot(
        x=loadings_x,
        y=loadings_y,
        data=loadings_df,
        hue=cazy_class,
        s=marker_size,
        style=ie_classifications,
    );
    
    g.axhline(0, linestyle='--', color='grey', linewidth=1.25);
    g.axvline(0, linestyle='--', color='grey', linewidth=1.25);
    g.set(xlim=(-1,1),ylim=(-1,1));
    plt.ylabel(f"PC{second_pc}") 
    plt.xlabel(f"PC{first_pc}")

    texts = [
        plt.text(
            xval,
            yval,
            lbl,
            ha='center',
            va='center',
            fontsize=font_size,
        ) for (xval, yval, lbl) in zip(
            loadings_x, loadings_y, loadings_labels
        ) if abs(xval) > threshold or abs(yval) > threshold
    ]
    adjustText.adjust_text(texts, arrowprops=dict(arrowstyle='-', color='black'));

    sns.move_legend(g, "lower center", bbox_to_anchor=(.5, 1), ncol=3, title=None, frameon=False);
    
    if file_path is not None:
        plt.savefig(file_path, dpi=dpi, bbox_inches='tight')
